"""
Provides physical, mathematical, computational, isotope, and material
constants, such as `Pi = 3.141592653589793238462643383279502884197`.

:copyright: Copyright 2012-2021 Matthias Cuntz, see AUTHORS.md for details.
:license: MIT License, see LICENSE for details.

Subpackages
===========
.. autosummary::
   const
"""
from .const import Pi, Pi2, Pi3, TwoPi, pi, pi2, pi3, Twopi, Sqrt2
from .const import Gravity, T0, P0, T25, sigma, R, R_air, R_H2O
from .const import Na, REarth, mmol_co2, mmol_h2o, mmol_air
from .const import density_quartz, cheat_quartz, cheat_water, cheat_air
from .const import latentheat_vaporization
from .const import R13VPDB, R18VSMOW, R2VSMOW
from .const import tiny, huge, eps


__all__ = [
    'Pi', 'Pi2', 'Pi3', 'TwoPi', 'pi', 'pi2', 'pi3', 'Twopi', 'Sqrt2',
    'Gravity', 'T0', 'P0', 'T25', 'sigma', 'R', 'R_air', 'R_H2O',
    'Na', 'REarth', 'mmol_co2', 'mmol_h2o', 'mmol_air',
    'density_quartz', 'cheat_quartz', 'cheat_water', 'cheat_air',
    'latentheat_vaporization',
    'R13VPDB', 'R18VSMOW', 'R2VSMOW',
    'tiny', 'huge', 'eps']



#!/usr/bin/env python
"""
This is the unittest for str2tex module.

python -m unittest -v tests/test_str2tex.py
python -m pytest --cov=pyjams --cov-report term-missing -v tests/test_str2tex.py

"""
from __future__ import division, absolute_import, print_function
import unittest


class TestStr2Tex(unittest.TestCase):
    """
    Tests for str2tex.py
    """

    def test_str2tex(self):
        import numpy as np
        from pyjams import str2tex

        # string
        strin = 'One'
        out = r'$\mathrm{One}$'
        assert str2tex(strin, usetex=True) == out
        assert str2tex(strin, bold=True, usetex=True) == (
            out.replace('rm', 'bf'))
        assert str2tex(strin, italic=True, usetex=True) == (
            out.replace('rm', 'it'))
        self.assertRaises(ValueError, str2tex, strin, bold=True, italic=True,
                          usetex=True)
        assert str2tex(strin, space2linebreak=True, usetex=True) == out
        assert str2tex(strin, usetex=False) == strin
        assert str2tex(strin, usetex=False, space2linebreak=True) == strin

        # list and -
        strin = ['One-', 'One-Two']
        out = [r'$\mathrm{One}$$\textrm{-}$',
               r'$\mathrm{One}$$\textrm{-}$$\mathrm{Two}$']
        self.assertEqual(str2tex(strin, usetex=True), out)
        outbf = [ i.replace('rm', 'bf') for i in out ]
        self.assertEqual(str2tex(strin, bold=True, usetex=True), outbf)
        outit = [ i.replace('rm', 'it') for i in out ]
        self.assertEqual(str2tex(strin, italic=True, usetex=True), outit)
        self.assertRaises(ValueError, str2tex, strin, bold=True, italic=True,
                          usetex=True)
        self.assertEqual(str2tex(strin, space2linebreak=True, usetex=True),
                         out)
        self.assertEqual(str2tex(strin, usetex=False), strin)
        self.assertEqual(str2tex(strin, usetex=False, space2linebreak=True),
                         strin)

        # tuple, space and linebreak
        strin = ('One Two', r'One\nTwo')
        out = (r'$\mathrm{One\ Two}$',
               r'$\mathrm{One}$\newline$\mathrm{Two}$')
        self.assertEqual(str2tex(strin, usetex=True), out)
        outbf = tuple([ i.replace('rm', 'bf') for i in out ])
        self.assertEqual(str2tex(strin, bold=True, usetex=True), outbf)
        outit = tuple([ i.replace('rm', 'it') for i in out ])
        self.assertEqual(str2tex(strin, italic=True, usetex=True), outit)
        self.assertRaises(ValueError, str2tex, strin, bold=True, italic=True,
                          usetex=True)
        outsp = tuple([ out[1] for i in out ])
        self.assertEqual(str2tex(strin, space2linebreak=True, usetex=True),
                         outsp)
        outn = tuple([ i.replace(r'\n', ''+'\n'+'') for i in strin ])
        self.assertEqual(str2tex(strin, usetex=False), outn)
        strinsp = tuple([ i.replace(' ', ''+'\n'+'') for i in outn ])
        self.assertEqual(str2tex(strin, usetex=False, space2linebreak=True),
                         strinsp)

        # ndarray incl. LaTeX
        strin = np.array([r'A $S_{Ti}$ is great\nbut use-less-'])
        out = np.array([r'$\mathrm{A\ }$$S_{Ti}$$\mathrm{\ is\ great}$\newline'
                        r'$\mathrm{but\ use}$$\textrm{-}$$\mathrm{less}$'
                        r'$\textrm{-}$'])
        self.assertEqual(str2tex(strin, usetex=True), out)
        outbf = np.array([ i.replace('rm', 'bf') for i in out ])
        self.assertEqual(str2tex(strin, bold=True, usetex=True), outbf)
        outit = np.array([ i.replace('rm', 'it') for i in out ])
        self.assertEqual(str2tex(strin, italic=True, usetex=True), outit)
        self.assertRaises(ValueError, str2tex, strin, bold=True, italic=True,
                          usetex=True)
        outsp = np.array([r'$\mathrm{A}$\newline$S_{Ti}$\newline$\mathrm{is}$'
                          r'\newline$\mathrm{great}$\newline$\mathrm{but}$'
                          r'\newline$\mathrm{use}$$\textrm{-}$$\mathrm{less}$'
                          r'$\textrm{-}$'])
        self.assertEqual(str2tex(strin, space2linebreak=True, usetex=True),
                         outsp)
        outn = tuple([ i.replace(r'\n', ''+'\n'+'') for i in strin ])
        self.assertEqual(str2tex(strin, usetex=False), outn)
        strinsp = tuple([ i.replace(' ', ''+'\n'+'') for i in outn ])
        self.assertEqual(str2tex(strin, usetex=False, space2linebreak=True),
                         strinsp)

        # complex string incl. LaTeX but use \n
        strin = r'$\alpha$Com_plex-str^ing\n(% of #string m$^{-2}$)$\alpha$'
        out = (r'$\alpha$$\mathrm{Com\_plex}$$\textrm{-}$$\mathrm{str\^ing}$'
               r'\newline$\mathrm{(\%\ of\ \#string\ m}$$^{-2}$$\mathrm{)}$'
               r'$\alpha$')
        assert str2tex(strin, usetex=True) == out
        assert str2tex(strin, bold=True, usetex=True) == (
            out.replace('rm', 'bf'))
        assert str2tex(strin, italic=True, usetex=True) == (
            out.replace('rm', 'it'))
        self.assertRaises(ValueError, str2tex, strin, bold=True, italic=True,
                          usetex=True)
        outsp = (r'$\alpha$$\mathrm{Com\_plex}$$\textrm{-}$$\mathrm{str\^ing}$'
                 r'\newline$\mathrm{(\%}$\newline$\mathrm{of}$\newline'
                 r'$\mathrm{\#string}$\newline$\mathrm{m}$$^{-2}$$\mathrm{)}$'
                 r'$\alpha$')
        assert str2tex(strin, space2linebreak=True, usetex=True) == outsp
        strinu = strin.replace(r'\n', ''+'\n'+'').replace('%', r'\%')
        self.assertEqual(str2tex(strin, usetex=False), strinu)
        strinsp = strinu.replace(' ', ''+'\n'+'')
        self.assertEqual(str2tex(strin, usetex=False, space2linebreak=True),
                         strinsp)

        # complex string incl. LaTeX
        strin = r'$\alpha$Com_plex-str^ing\newline(% of #string m$^{-2}$)$\alpha$'
        out = (r'$\alpha$$\mathrm{Com\_plex}$$\textrm{-}$$\mathrm{str\^ing}$'
               r'\newline$\mathrm{(\%\ of\ \#string\ m}$$^{-2}$$\mathrm{)}$'
               r'$\alpha$')
        assert str2tex(strin, usetex=True) == out
        assert str2tex(strin, bold=True, usetex=True) == (
            out.replace('rm', 'bf'))
        assert str2tex(strin, italic=True, usetex=True) == (
            out.replace('rm', 'it'))
        self.assertRaises(ValueError, str2tex, strin, bold=True, italic=True,
                          usetex=True)
        outsp = (r'$\alpha$$\mathrm{Com\_plex}$$\textrm{-}$$\mathrm{str\^ing}$'
                 r'\newline$\mathrm{(\%}$\newline$\mathrm{of}$\newline'
                 r'$\mathrm{\#string}$\newline$\mathrm{m}$$^{-2}$$\mathrm{)}$'
                 r'$\alpha$')
        assert str2tex(strin, space2linebreak=True, usetex=True) == outsp
        strinu = strin.replace(r'\newline', ''+'\n'+'').replace('%', r'\%')
        self.assertEqual(str2tex(strin, usetex=False), strinu)
        strinsp = strinu.replace(' ', ''+'\n'+'')
        self.assertEqual(str2tex(strin, usetex=False, space2linebreak=True),
                         strinsp)

        # complex string no LaTeX no raw string
        strin = 'Com_plex-str^ing (% of #string m)'
        out = (r'$\mathrm{Com\_plex}$$\textrm{-}$$\mathrm{str\^ing'
               r'\ (\%\ of\ \#string\ m)}$')
        assert str2tex(strin, usetex=True) == out
        assert str2tex(strin, bold=True, usetex=True) == (
            out.replace('rm', 'bf'))
        assert str2tex(strin, italic=True, usetex=True) == (
            out.replace('rm', 'it'))
        self.assertRaises(ValueError, str2tex, strin, bold=True, italic=True,
                          usetex=True)
        outsp = (r'$\mathrm{Com\_plex}$$\textrm{-}$$\mathrm{str\^ing}$'
                 r'\newline$\mathrm{(\%}$\newline$\mathrm{of}$\newline$\mathrm{\#string}$'
                 r'\newline$\mathrm{m)}$')
        assert str2tex(strin, space2linebreak=True, usetex=True) == outsp
        strinu = strin.replace(r'\n', ''+'\n'+'').replace('%', r'\%')
        assert str2tex(strin, usetex=False) == strinu
        strinsp = strinu.replace(' ', ''+'\n'+'')
        assert str2tex(strin, usetex=False, space2linebreak=True) == strinsp

        # complex string no LaTeX using \n
        strin = r'Com_plex-str^ing\n(% of #string m)-'
        out = (r'$\mathrm{Com\_plex}$$\textrm{-}$$\mathrm{str\^ing}$'
               r'\newline$\mathrm{(\%\ of\ \#string\ m)}$$\textrm{-}$')
        assert str2tex(strin, usetex=True) == out
        assert str2tex(strin, bold=True, usetex=True) == (
            out.replace('rm', 'bf'))
        assert str2tex(strin, italic=True, usetex=True) == (
            out.replace('rm', 'it'))
        self.assertRaises(ValueError, str2tex, strin, bold=True, italic=True,
                          usetex=True)
        outsp = (r'$\mathrm{Com\_plex}$$\textrm{-}$$\mathrm{str\^ing}$'
                 r'\newline$\mathrm{(\%}$\newline$\mathrm{of}$\newline'
                 r'$\mathrm{\#string}$\newline$\mathrm{m)}$$\textrm{-}$')
        assert str2tex(strin, space2linebreak=True, usetex=True) == outsp
        strinu = strin.replace(r'\n', ''+'\n'+'').replace('%', r'\%')
        assert str2tex(strin, usetex=False) == strinu
        strinsp = strinu.replace(' ', ''+'\n'+'')
        assert str2tex(strin, usetex=False, space2linebreak=True) == strinsp

        # complex string no LaTeX using \newline
        strin = r'Com_plex-str^ing\newline(% of #string m)-'
        out = (r'$\mathrm{Com\_plex}$$\textrm{-}$$\mathrm{str\^ing}$'
               r'\newline$\mathrm{(\%\ of\ \#string\ m)}$$\textrm{-}$')
        assert str2tex(strin, usetex=True) == out
        assert str2tex(strin, bold=True, usetex=True) == (
            out.replace('rm', 'bf'))
        assert str2tex(strin, italic=True, usetex=True) == (
            out.replace('rm', 'it'))
        self.assertRaises(ValueError, str2tex, strin, bold=True, italic=True,
                          usetex=True)
        outsp = (r'$\mathrm{Com\_plex}$$\textrm{-}$$\mathrm{str\^ing}$'
                 r'\newline$\mathrm{(\%}$\newline$\mathrm{of}$\newline'
                 r'$\mathrm{\#string}$\newline$\mathrm{m)}$$\textrm{-}$')
        assert str2tex(strin, space2linebreak=True, usetex=True) == outsp
        strinu = strin.replace(r'\newline', ''+'\n'+'').replace('%', r'\%')
        assert str2tex(strin, usetex=False) == strinu
        strinsp = strinu.replace(' ', ''+'\n'+'')
        assert str2tex(strin, usetex=False, space2linebreak=True) == strinsp


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
"""
pyjams is a general Python package with miscellaneous utility functions used in
several other packages. It is offering miscellaneous functions in different
categories, such as reading different file formats, julian date routines, or
meteorological functions. It has several subpackages offering constants or
special functions as well as objective functions to be used with
scipy.optimize.fmin or scipy.optimize.curvefit, and much more.

The package has evolved from its predecessor the JAMS Python package
https://github.com/mcuntz/jams_python

:copyright: Copyright 2021-2022 Matthias Cuntz, see AUTHORS.md for details.
:license: MIT License, see LICENSE for details.

Subpackages
===========
.. autosummary::
   alpha_equ_h2o
   alpha_kin_h2o
   argsort
   gridcellarea
   closest
   color
   const
   date2date
   division
   datetime
   esat
   fgui
   fsread
   functions
   kernel_regression
   mad
   mcplot
   morris_method
   ncio
   position
   readnetcdf
   romanliterals
   screening
   str2tex
   tee
   text2plot

History
    * Written Oct 2021 by Matthias Cuntz (mc (at) macu (dot) de)
    * v1.0, initial Github, PyPI, Zenodo commit, Oct 2021, Matthias Cuntz
    * v1.1, automatic versioning, zenodo defaults, Oct 2021, Matthias Cuntz
    * v1.2, added closest, Oct 2021, Matthias Cuntz
    * v1.3, added argsort, Oct 2021, Matthias Cuntz
    * v1.4, added division, Oct 2021, Matthias Cuntz
    * v1.5, added alpha_equ_h2o, Oct 2021, Matthias Cuntz
    * v1.6, added alpha_kin_h2o, Nov 2021, Matthias Cuntz
    * v1.7, added mcPlot, Nov 2021, Matthias Cuntz
    * v1.8, added str2tex and color, Nov 2021, Matthias Cuntz
    * v1.9, added position, Nov 2021, Matthias Cuntz
    * v1.10, added sron colors, Nov 2021, Matthias Cuntz
    * v1.11, added text2plot, abc2plot, signature2plot,
      Nov 2021, Matthias Cuntz
    * v1.12, added date2dec and all its wrappers, Dec 2021, Matthias Cuntz
    * v1.13, added fsread, fread, sread, Dec 2021, Matthias Cuntz
    * v1.14, added GUI dialogs to choose files and directories using Tkinter,
      Jan 2022, Matthias Cuntz
    * v1.15, added esat, Jan 2022, Matthias Cuntz
    * v1.16, added mad, Jan 2022, Matthias Cuntz
    * v1.17, added xread, xlsread, and xlsxread, Jan 2022, Matthias Cuntz
    * v1.18, helper module to preserve input types, Mar 2022, Matthias Cuntz
    * v1.19, added readnetcdf, Mar 2022, Matthias Cuntz
    * v1.20, added gridcellarea and kernel_regression, Apr 2022, Matthias Cuntz
    * v1.21, enhancements in color and kernel_regression,
      Apr 2022, Matthias Cuntz
    * v1.22, added ncio, May 2022, Matthias Cuntz
    * v1.23, added datetime module, Jun 2022, Matthias Cuntz

"""
# version, author
try:  # pragma: no cover
    from ._version import __version__
except ImportError:  # pragma: no cover
    # package is not installed
    __version__ = "0.0.0.dev0"
__author__  = "Matthias Cuntz, Juliane Mai, Stephan Thober, Arndt Piayda"

# sub-packages without dependencies to rest of pyjams
# color palettes and continuous color maps
from . import color
# physical, mathematical, computational, isotope, and material constants
from . import const
# variety of specialised functions
from . import functions
# netCDF4 functions to copy netcdf file while doing some transformations on
# variables and dimensions.
from . import ncio

# isotopic fractionation factors during liquid-water vapour equilibration
from .alpha_equ_h2o import alpha_equ_h2o
# kinetic fractionation of molecular diffusion of water vapour
from .alpha_kin_h2o import alpha_kin_h2o
# argmax, argmin and argsort for array_like and Python iterables
from .argsort import argmax, argmin, argsort
# Area of grid cells on Earth
from .gridcellarea import gridcellarea
# (index of) closest element in an array
from .closest import closest
# convert date representations between different regional variants
from .date2date import date2date
from .date2date import date2en, date2fr, date2us
from .date2date import en2date, en2fr, en2us
from .date2date import fr2date, fr2en, fr2us
from .date2date import us2date, us2en, us2fr
# catch division by zero
from .division import division, div
# cftime extension
from .datetime import date2dec, date2num, dec2date, num2date, datetime
# saturation vapour pressure over water and ice
from .esat import esat
# GUI dialogs to choose files and directories using Tkinter
from .fgui import directory_from_gui, directories_from_gui
from .fgui import file_from_gui, files_from_gui
# read numbers and strings from a file into 2D float and string arrays
from .fsread import fsread, fread, sread
from .fsread import xread, xlsread, xlsxread
# multi-dimensional non-parametric kernel regression
from .kernel_regression import kernel_regression_h, kernel_regression
# median absolute deviation test
from .mad import mad
# Matthias Cuntz' standard plotting class.
from .mcplot import mcPlot
# has to be ordered for import: morris -> screening
# Sampling of optimised trajectories for and calculation of Morris Measures /
# Elementary Effects
from .morris_method import morris_sampling, elementary_effects
# positions of subplots, used with add_axes
from .position import position
# get variables from or print information of a netcdf file
from .readnetcdf import infonetcdf, ncinfo, readnetcdf, ncread
# Convert integer to and from Roman numerals
from .romanliterals import int2roman, roman2int
# Sample trajectories, run model and return Morris Elementary Effects
from .screening import screening, ee
# Convert strings to LaTeX strings
from .str2tex import str2tex
# like unix tee
from .tee import tee
# put text on plot (import after str2tex)
from .text2plot import text2plot, abc2plot, signature2plot


__all__ = ['__version__', '__author__',
           'color', 'const', 'functions',
           'alpha_equ_h2o', 'alpha_kin_h2o',
           'argmax', 'argmin', 'argsort',
           'gridcellarea',
           'closest',
           'date2date',
           'date2en', 'date2fr', 'date2us',
           'en2date', 'en2fr', 'en2us',
           'fr2date', 'fr2en', 'fr2us',
           'us2date', 'us2en', 'us2fr',
           'division', 'div',
           'date2dec', 'date2num', 'dec2date', 'num2date', 'datetime',
           'esat',
           'directory_from_gui', 'directories_from_gui',
           'file_from_gui', 'files_from_gui',
           'fsread', 'fread', 'sread',
           'xread', 'xlsread', 'xlsxread',
           'kernel_regression_h', 'kernel_regression'
           'mad',
           'mcPlot',
           'morris_sampling', 'elementary_effects',
           'position',
           'infonetcdf', 'ncinfo', 'readnetcdf', 'ncread',
           'int2roman', 'roman2int',
           'screening', 'ee',
           'str2tex',
           'tee',
           'text2plot', 'abc2plot', 'signature2plot',
           ]



#!/usr/bin/env python
"""
This is the unittest for the Morris module.

python -m unittest -v tests/test_morris_method.py
python -m pytest --cov=pyjams --cov-report term-missing -v tests/test_morris_method.py

"""
from __future__ import division, absolute_import, print_function
import unittest


class TestMorrisMethod(unittest.TestCase):
    """
    Tests for morris_method.py
    Missing coverage:
        ca. 405ff: exactly same trajectory sampled by chance
        ca. 500-550: ImportError if matplotlib not installed
        ca. 730: only one trajectory

    """
    def setUp(self):
        import numpy as np
        # seed for reproducible results
        self.seed = 1234
        np.random.seed(seed=self.seed)
        self.nparam     = 15
        self.LB         = np.arange(self.nparam)
        self.UB         = 2. * self.LB + 1.
        self.Diagnostic = 0

    def test_r_10(self):
        import numpy as np
        from pyjams import morris_sampling, elementary_effects

        nt     = 10
        nsteps = 6
        ntotal = 100
        out = np.random.random(nt*(self.nparam+1))

        # Check 1
        mat, vec = morris_sampling(self.nparam, self.LB, self.UB, nt,
                                   nsteps=nsteps, ntotal=ntotal,
                                   Diagnostic=self.Diagnostic)
        self.assertEqual(list(np.around(mat[0, 0:5], 3)),
                         [0.6, 2.2, 2., 4.6, 8.])
        self.assertEqual(list(np.around(vec[0:5], 3)),
                         [12., 11., 5., 1., 9.])

        # Check 2
        sa, res = elementary_effects(self.nparam, mat, vec, out, nsteps=nsteps)
        self.assertEqual(list(np.around(res[0:5, 0], 3)),
                         [0.485, 0.445, 0.438, 0.58, 0.645])
        self.assertEqual(list(np.around(sa[0:5, 1], 3)),
                         [0.47, 0.502, 0.816, 0.722, 0.418])

    def test_r_10_total(self):
        import numpy as np
        from pyjams import morris_sampling, elementary_effects

        nt     = 10
        nsteps = 6
        ntotal = None
        out = np.random.random(nt*(self.nparam+1))

        # Check 1
        mat, vec = morris_sampling(self.nparam, self.LB, self.UB, nt,
                                   nsteps=nsteps, ntotal=ntotal,
                                   Diagnostic=self.Diagnostic)
        self.assertEqual(list(np.around(mat[0, 0:5], 3)),
                         [0.6, 2.2, 2., 4.6, 8.])
        self.assertEqual(list(np.around(vec[0:5], 3)),
                         [12., 11., 5., 1., 9.])

        # Check 2
        sa, res = elementary_effects(self.nparam, mat, vec, out, nsteps=nsteps)
        self.assertEqual(list(np.around(res[0:5, 0], 3)),
                         [0.485, 0.445, 0.438, 0.58, 0.645])
        self.assertEqual(list(np.around(sa[0:5, 1], 3)),
                         [0.47, 0.502, 0.816, 0.722, 0.418])

    def test_r_10_diag(self):
        import os
        import numpy as np
        from pyjams import morris_sampling, elementary_effects

        nt     = 10
        nsteps = 6
        ntotal = 100
        out = np.random.random(nt*(self.nparam+1))

        # Check 1
        mat, vec = morris_sampling(self.nparam, self.LB, self.UB, nt,
                                   nsteps=nsteps, ntotal=ntotal, Diagnostic=1)
        self.assertEqual(list(np.around(mat[0, 0:5], 3)),
                         [0.6, 2.2, 2., 4.6, 8.])
        self.assertEqual(list(np.around(vec[0:5], 3)),
                         [12., 11., 5., 1., 9.])
        self.assertTrue(os.path.exists('morris_diag_new_strategy.png'))
        self.assertTrue(os.path.exists('morris_diag_old_strategy.png'))
        if os.path.exists('morris_diag_new_strategy.png'):
            os.remove('morris_diag_new_strategy.png')
        if os.path.exists('morris_diag_old_strategy.png'):
            os.remove('morris_diag_old_strategy.png')

        # Check 2
        sa, res = elementary_effects(self.nparam, mat, vec, out, nsteps=nsteps,
                                     Diagnostic=1)
        self.assertEqual(list(np.around(res[0:5, 0], 3)),
                         [0.485, 0.445, 0.438, 0.58, 0.645])
        self.assertEqual(list(np.around(sa[0:5, 1], 3)),
                         [0.47, 0.502, 0.816, 0.722, 0.418])
        self.assertFalse(os.path.exists('morris_diag_new_strategy.png'))
        self.assertFalse(os.path.exists('morris_diag_old_strategy.png'))

    def test_r_10_nan(self):
        import numpy as np
        from pyjams import morris_sampling, elementary_effects

        nt     = 10
        nsteps = 6
        ntotal = 100
        out = np.random.random(nt*(self.nparam+1))
        out[1:nt*self.nparam:self.nparam//2] = np.nan

        mat, vec = morris_sampling(self.nparam, self.LB, self.UB, nt,
                                   nsteps=nsteps, ntotal=ntotal,
                                   Diagnostic=self.Diagnostic)
        sa, res = elementary_effects(self.nparam, mat, vec, out, nsteps=nsteps)
        self.assertEqual(list(np.around(res[0:5, 0], 3)),
                         [0.368, 0.309, 0.549, 0.534, 0.65])
        self.assertEqual(list(np.around(sa[~np.isnan(sa[:, 1]), 1], 3)),
                         [0.47, 0.816, 0.722, 0.418, -0.653, -0.941, 0.863,
                          -1.265, -0.424, -0.786, 0.183])

    def test_r_1(self):
        import numpy as np
        from pyjams import morris_sampling, elementary_effects

        nt     = 1
        nsteps = 6
        ntotal = 100
        out = np.random.random(nt*(self.nparam+1))

        mat, vec = morris_sampling(self.nparam, self.LB, self.UB, nt,
                                   nsteps=nsteps, ntotal=ntotal,
                                   Diagnostic=self.Diagnostic)
        sa, res = elementary_effects(self.nparam, mat, vec, out, nsteps=nsteps)
        self.assertEqual(list(np.around(res[0:5, 0], 3)),
                         [0.579, 0.009, 0.239, 0.864, 0.876])
        self.assertEqual(list(np.around(sa[0:5].squeeze(), 3)),
                         [-0.579, -0.009, -0.239, -0.864, 0.876])

    def test_groups(self):
        import numpy as np
        from pyjams import morris_sampling, elementary_effects

        ngroup = 5
        Groups = np.random.randint(0, 4, (self.nparam, ngroup))
        nt     = 10
        nsteps = 6
        ntotal = 100
        out = np.random.random(nt*(self.nparam+1))

        # Check 1
        mat, vec = morris_sampling(self.nparam, self.LB, self.UB, nt,
                                   nsteps=nsteps, ntotal=ntotal,
                                   GroupMat=Groups, Diagnostic=self.Diagnostic)
        self.assertEqual(list(np.around(mat[0, 0:5], 3)),
                         [0.2, 1.8, 3.8, 7., 8.])
        self.assertEqual(list(np.around(vec[0:5], 3)),
                         [3., 0., 1., 4., 2.])

        # Check 2
        sa, res = elementary_effects(self.nparam, mat, vec, out, nsteps=nsteps,
                                     Group=Groups)
        self.assertEqual(list(np.around(res[0:5, 0], 3)),
                         [0.531, 0.43, 0.432, 0.443, 0.443])
        self.assertEqual(list(np.around(sa[0:5, 1], 3)),
                         [0.279, 0.557, 0.557, 0.557, 0.557])

    def test_groups_diagnostics(self):
        import numpy as np
        from pyjams import morris_sampling, elementary_effects

        ngroup = 5
        Groups = np.random.randint(0, 4, (self.nparam, ngroup))
        nt     = 10
        nsteps = 6
        ntotal = 100
        out = np.random.random(nt*(self.nparam+1))

        # Check 1
        mat, vec = morris_sampling(self.nparam, self.LB, self.UB, nt,
                                   nsteps=nsteps, ntotal=ntotal,
                                   GroupMat=Groups, Diagnostic=1)
        self.assertEqual(list(np.around(mat[0, 0:5], 3)),
                         [0.2, 1.8, 3.8, 7., 8.])
        self.assertEqual(list(np.around(vec[0:5], 3)),
                         [3., 0., 1., 4., 2.])

        # Check 2
        sa, res = elementary_effects(self.nparam, mat, vec, out, nsteps=nsteps,
                                     Group=Groups)
        self.assertEqual(list(np.around(res[0:5, 0], 3)),
                         [0.531, 0.43, 0.432, 0.443, 0.443])
        self.assertEqual(list(np.around(sa[0:5, 1], 3)),
                         [0.279, 0.557, 0.557, 0.557, 0.557])

    def test_groups_nan(self):
        import numpy as np
        from pyjams import morris_sampling, elementary_effects

        ngroup = 5
        Groups = np.random.randint(0, 4, (self.nparam, ngroup))
        nt     = 10
        nsteps = 6
        ntotal = 100
        out = np.random.random(nt*(self.nparam+1))
        out[1:nt*self.nparam:self.nparam//2] = np.nan

        mat, vec = morris_sampling(self.nparam, self.LB, self.UB, nt,
                                   nsteps=nsteps, ntotal=ntotal,
                                   GroupMat=Groups, Diagnostic=self.Diagnostic)
        sa, res = elementary_effects(self.nparam, mat, vec, out,
                                     nsteps=nsteps, Group=Groups)
        self.assertEqual(list(np.around(res[0:5, 0], 3)),
                         [0.49, 0.425, 0.427, 0.441, 0.441])
        self.assertEqual(list(np.around(sa[0:5, 1], 3)),
                         [0.279, 0.557, 0.557, 0.557, 0.557])

    def test_r_10_dist(self):
        import numpy as np
        import scipy.stats as stats
        from pyjams import morris_sampling, elementary_effects

        nt     = 10
        nsteps = 6
        ntotal = 100
        out = np.random.random(nt*(self.nparam+1))

        dist      = [ stats.uniform for i in range(self.nparam) ]
        distparam = [ (self.LB[i], self.UB[i]-self.LB[i])
                      for i in range(self.nparam) ]
        LB = np.zeros(self.nparam)
        UB = np.ones(self.nparam)

        # Check 1
        mat, vec = morris_sampling(self.nparam, LB, UB, nt, nsteps=nsteps,
                                   ntotal=ntotal, dist=dist,
                                   distparam=distparam,
                                   Diagnostic=self.Diagnostic)
        self.assertEqual(list(np.around(mat[0, 0:5], 3)),
                         [0.6, 2.2, 2., 4.6, 8.])
        self.assertEqual(list(np.around(vec[0:5], 3)),
                         [12., 11., 5., 1., 9.])

        # Check 2
        sa, res = elementary_effects(self.nparam, mat, vec, out, nsteps=nsteps)
        self.assertEqual(list(np.around(res[0:5, 0], 3)),
                         [0.485, 0.445, 0.438, 0.58, 0.645])
        self.assertEqual(list(np.around(sa[0:5, 1], 3)),
                         [0.47, 0.502, 0.816, 0.722, 0.418])

    def test_r_10_dist_none(self):
        import numpy as np
        import scipy.stats as stats
        from pyjams import morris_sampling, elementary_effects

        nt     = 10
        nsteps = 6
        ntotal = 100
        out = np.random.random(nt*(self.nparam+1))

        dist      = [ stats.uniform if i%2 == 0 else None
                      for i in range(self.nparam) ]
        distparam = [ (self.LB[i], self.UB[i]-self.LB[i])
                      for i in range(self.nparam) ]
        LB = self.LB
        UB = self.UB
        LB[::2] = 0.
        UB[::2] = 1.

        # Check 1
        mat, vec = morris_sampling(self.nparam, LB, UB, nt, nsteps=nsteps,
                                   ntotal=ntotal, dist=dist,
                                   distparam=distparam,
                                   Diagnostic=self.Diagnostic)
        self.assertEqual(list(np.around(mat[0, 0:5], 3)),
                         [0.6, 2.2, 2., 4.6, 8.])
        self.assertEqual(list(np.around(vec[0:5], 3)),
                         [12., 11., 5., 1., 9.])

        # Check 2
        sa, res = elementary_effects(self.nparam, mat, vec, out, nsteps=nsteps)
        self.assertEqual(list(np.around(res[0:5, 0], 3)),
                         [0.485, 0.445, 0.438, 0.58, 0.645])
        self.assertEqual(list(np.around(sa[0:5, 1], 3)),
                         [0.47, 0.502, 0.816, 0.722, 0.418])

        LB = np.zeros(self.nparam)
        UB = np.ones(self.nparam)
        dist       = [ stats.uniform for i in range(self.nparam) ]
        distparam1 = [ (LB[i], UB[i]-LB[i]) for i in range(self.nparam) ]
        distparam2 = [ (LB[i], UB[i]-LB[i]) if i%2 == 0 else None
                       for i in range(self.nparam) ]

        # Check 3
        np.random.seed(seed=self.seed)
        mat1, vec1 = morris_sampling(self.nparam, LB, UB, nt, nsteps=nsteps,
                                     ntotal=ntotal, dist=dist,
                                     distparam=distparam1,
                                     Diagnostic=self.Diagnostic)
        np.random.seed(seed=self.seed)
        mat2, vec2 = morris_sampling(self.nparam, LB, UB, nt, nsteps=nsteps,
                                     ntotal=ntotal, dist=dist,
                                     distparam=distparam2,
                                     Diagnostic=self.Diagnostic)
        self.assertTrue(np.all((mat1-mat2) == 0.))
        self.assertTrue(np.all((vec1-vec2) == 0.))

        # Check 3
        np.random.seed(seed=self.seed)
        mat3, vec3 = morris_sampling(self.nparam, LB, UB, nt, nsteps=nsteps,
                                     ntotal=ntotal, dist=dist, distparam=None,
                                     Diagnostic=self.Diagnostic)
        self.assertTrue(np.all((mat1-mat3) == 0.))
        self.assertTrue(np.all((vec1-vec3) == 0.))

    def test_r_10_dist_raise(self):
        import numpy as np
        import scipy.stats as stats
        from pyjams import morris_sampling

        nt     = 10
        nsteps = 6
        ntotal = 100
        out = np.random.random(nt*(self.nparam+1))

        dist      = [ stats.uniform for i in range(self.nparam) ]
        distparam = [ (self.LB[i], self.UB[i]-self.LB[i])
                      for i in range(self.nparam) ]
        LB = np.zeros(self.nparam)
        UB = np.ones(self.nparam)

        # Check 1
        self.assertRaises(AssertionError, morris_sampling, self.nparam, LB[1:],
                          UB, nt, nsteps=nsteps, ntotal=ntotal,
                          Diagnostic=self.Diagnostic)

        # Check 2
        self.assertRaises(AssertionError, morris_sampling, self.nparam, LB,
                          UB[1:], nt, nsteps=nsteps, ntotal=ntotal,
                          Diagnostic=self.Diagnostic)

        # Check 3
        self.assertRaises(AssertionError, morris_sampling, self.nparam, LB, UB,
                          nt, nsteps=nsteps, ntotal=ntotal, dist=dist[1:],
                          distparam=distparam, Diagnostic=self.Diagnostic)

        # Check 4
        dist[0] = 'stats.uniform'
        self.assertRaises(TypeError, morris_sampling, self.nparam, LB, UB, nt,
                          nsteps=nsteps, ntotal=ntotal,
                          dist=dist, distparam=distparam,
                          Diagnostic=self.Diagnostic)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
"""
This is the unittest for readnetcdf module.

python -m unittest -v tests/test_readnetcdf.py
python -m pytest --cov=pyjams --cov-report term-missing -v tests/test_readnetcdf.py

"""
import unittest


def _flatten(itr):
    import numpy as np
    fitr = np.array(itr).flatten()
    if len(fitr) == 0:
        return list(fitr)
    else:
        if isinstance(fitr[0], str):
            return [ i for i in fitr ]
        else:
            return [ i if np.isfinite(i) else np.finfo(float).max
                     for i in fitr ]


class TestInfonetcdf(unittest.TestCase):
    """
    Tests for readnetcdf.py
    """

    def test_infonetcdf(self):
        import numpy as np
        import netCDF4 as nc
        from pyjams import infonetcdf

        ncfile = 'tests/test_readnetcdf.nc'

        # variables
        fout = infonetcdf(ncfile, variables=True)
        fsoll = ['x', 'y', 'is1', 'is2']
        assert isinstance(fout, list)
        self.assertEqual(fout, fsoll)

        # variables, sort
        fout = infonetcdf(ncfile, variables=True, sort=True)
        fsoll = ['is1', 'is2', 'x', 'y']
        assert isinstance(fout, list)
        self.assertEqual(fout, fsoll)

        # codes
        fout = infonetcdf(ncfile, codes=True)
        fsoll = [-1, -1, 128, 129]
        assert isinstance(fout, list)
        self.assertEqual(fout, fsoll)

        # codes, sort
        fout = infonetcdf(ncfile, codes=True, sort=True)
        fsoll = [128, 129, -1, -1]
        assert isinstance(fout, list)
        self.assertEqual(fout, fsoll)

        # units
        fout = infonetcdf(ncfile, units=True)
        fsoll = ['xx', 'yy', 'arbitrary', 'arbitrary']
        assert isinstance(fout, list)
        self.assertEqual(fout, fsoll)

        # units, sort
        fout = infonetcdf(ncfile, units=True, sort=True)
        fsoll = ['arbitrary', 'arbitrary', 'xx', 'yy']
        assert isinstance(fout, list)
        self.assertEqual(fout, fsoll)

        # long_names
        fout = infonetcdf(ncfile, long_names=True)
        fsoll = ['x-axis', 'y-axis', 'all ones', 'all twos']
        assert isinstance(fout, list)
        self.assertEqual(fout, fsoll)

        # long_names, sort
        fout = infonetcdf(ncfile, long_names=True, sort=True)
        fsoll = ['all ones', 'all twos', 'x-axis', 'y-axis']
        assert isinstance(fout, list)
        self.assertEqual(fout, fsoll)

        # dims, var
        fout = infonetcdf(ncfile, var='is1', dims=True)
        fsoll = ('y', 'x')
        assert isinstance(fout, tuple)
        self.assertEqual(fout, fsoll)

        # dims, code
        fout = infonetcdf(ncfile, code=128, dims=True)
        fsoll = ('y', 'x')
        assert isinstance(fout, tuple)
        self.assertEqual(fout, fsoll)

        # shape, var
        fout = infonetcdf(ncfile, var='is1', shape=True)
        fsoll = (2, 4)
        assert isinstance(fout, tuple)
        self.assertEqual(fout, fsoll)

        # shape, code
        fout = infonetcdf(ncfile, code=128, shape=True)
        fsoll = (2, 4)
        assert isinstance(fout, tuple)
        self.assertEqual(fout, fsoll)

        # variable attributes, var
        fout = infonetcdf(ncfile, var='is1', attributes=True)
        fsoll = ['code', 'long_name', 'units']
        assert isinstance(fout, dict)
        fout = sorted(fout.keys())
        fsoll = sorted(fsoll)
        self.assertEqual(fout, fsoll)

        # variable attributes, var
        fout = infonetcdf(ncfile, 'is1', attributes=True)
        fsoll = ['code', 'long_name', 'units']
        assert isinstance(fout, dict)
        fout = sorted(fout.keys())
        fsoll = sorted(fsoll)
        self.assertEqual(fout, fsoll)

        # variable attributes, code
        fout = infonetcdf(ncfile, code=128, attributes=True)
        fsoll = ['code', 'long_name', 'units']
        assert isinstance(fout, dict)
        fout = sorted(fout.keys())
        fsoll = sorted(fsoll)
        self.assertEqual(fout, fsoll)

        # file attributes
        fout = infonetcdf(ncfile, attributes=True)
        fsoll = ['creator', 'history', 'NCO']
        assert isinstance(fout, dict)
        fout = sorted(fout.keys())
        fsoll = sorted(fsoll)
        self.assertEqual(fout, fsoll)

        # no keyword
        fout = infonetcdf(ncfile)
        fsoll = None
        assert fout is fsoll

        # errors
        # dims, no var
        self.assertRaises(ValueError, infonetcdf, ncfile, dims=True)
        # dims, var does not exist
        self.assertRaises(ValueError, infonetcdf, ncfile, dims=True,
                          var='is3')
        # dims, code does not exist
        self.assertRaises(ValueError, infonetcdf, ncfile, dims=True,
                          code=130)
        # shape, no var
        self.assertRaises(ValueError, infonetcdf, ncfile, shape=True)
        # shape, var does not exist
        self.assertRaises(ValueError, infonetcdf, ncfile, shape=True,
                          var='is3')
        # shape, code does not exist
        self.assertRaises(ValueError, infonetcdf, ncfile, shape=True,
                          code=130)
        # attributes, var does not exist
        self.assertRaises(ValueError, infonetcdf, ncfile, attributes=True,
                          var='is3')
        # attributes, code does not exist
        self.assertRaises(ValueError, infonetcdf, ncfile, attributes=True,
                          code=130)

    def test_ncinfo(self):
        import numpy as np
        import netCDF4 as nc
        from pyjams import ncinfo

        ncfile = 'tests/test_readnetcdf.nc'

        # variables
        fout = ncinfo(ncfile, variables=True)
        fsoll = ['x', 'y', 'is1', 'is2']
        assert isinstance(fout, list)
        self.assertEqual(fout, fsoll)

        # variables, sort
        fout = ncinfo(ncfile, variables=True, sort=True)
        fsoll = ['is1', 'is2', 'x', 'y']
        assert isinstance(fout, list)
        self.assertEqual(fout, fsoll)

        # codes
        fout = ncinfo(ncfile, codes=True)
        fsoll = [-1, -1, 128, 129]
        assert isinstance(fout, list)
        self.assertEqual(fout, fsoll)

        # codes, sort
        fout = ncinfo(ncfile, codes=True, sort=True)
        fsoll = [128, 129, -1, -1]
        assert isinstance(fout, list)
        self.assertEqual(fout, fsoll)

        # units
        fout = ncinfo(ncfile, units=True)
        fsoll = ['xx', 'yy', 'arbitrary', 'arbitrary']
        assert isinstance(fout, list)
        self.assertEqual(fout, fsoll)

        # units, sort
        fout = ncinfo(ncfile, units=True, sort=True)
        fsoll = ['arbitrary', 'arbitrary', 'xx', 'yy']
        assert isinstance(fout, list)
        self.assertEqual(fout, fsoll)

        # long_names
        fout = ncinfo(ncfile, long_names=True)
        fsoll = ['x-axis', 'y-axis', 'all ones', 'all twos']
        assert isinstance(fout, list)
        self.assertEqual(fout, fsoll)

        # long_names, sort
        fout = ncinfo(ncfile, long_names=True, sort=True)
        fsoll = ['all ones', 'all twos', 'x-axis', 'y-axis']
        assert isinstance(fout, list)
        self.assertEqual(fout, fsoll)

        # dims, var
        fout = ncinfo(ncfile, var='is1', dims=True)
        fsoll = ('y', 'x')
        assert isinstance(fout, tuple)
        self.assertEqual(fout, fsoll)

        # dims, code
        fout = ncinfo(ncfile, code=128, dims=True)
        fsoll = ('y', 'x')
        assert isinstance(fout, tuple)
        self.assertEqual(fout, fsoll)

        # shape, var
        fout = ncinfo(ncfile, var='is1', shape=True)
        fsoll = (2, 4)
        assert isinstance(fout, tuple)
        self.assertEqual(fout, fsoll)

        # shape, code
        fout = ncinfo(ncfile, code=128, shape=True)
        fsoll = (2, 4)
        assert isinstance(fout, tuple)
        self.assertEqual(fout, fsoll)

        # variable attributes, var
        fout = ncinfo(ncfile, var='is1', attributes=True)
        fsoll = ['code', 'long_name', 'units']
        assert isinstance(fout, dict)
        fout = sorted(fout.keys())
        fsoll = sorted(fsoll)
        self.assertEqual(fout, fsoll)

        # variable attributes, var
        fout = ncinfo(ncfile, 'is1', attributes=True)
        fsoll = ['code', 'long_name', 'units']
        assert isinstance(fout, dict)
        fout = sorted(fout.keys())
        fsoll = sorted(fsoll)
        self.assertEqual(fout, fsoll)

        # variable attributes, code
        fout = ncinfo(ncfile, code=128, attributes=True)
        fsoll = ['code', 'long_name', 'units']
        assert isinstance(fout, dict)
        fout = sorted(fout.keys())
        fsoll = sorted(fsoll)
        self.assertEqual(fout, fsoll)

        # file attributes
        fout = ncinfo(ncfile, attributes=True)
        fsoll = ['creator', 'history', 'NCO']
        assert isinstance(fout, dict)
        fout = sorted(fout.keys())
        fsoll = sorted(fsoll)
        self.assertEqual(fout, fsoll)

        # no keyword
        fout = ncinfo(ncfile)
        fsoll = None
        assert fout is fsoll

        # errors
        # dims, no var
        self.assertRaises(ValueError, ncinfo, ncfile, dims=True)
        # dims, var does not exist
        self.assertRaises(ValueError, ncinfo, ncfile, dims=True,
                          var='is3')
        # dims, code does not exist
        self.assertRaises(ValueError, ncinfo, ncfile, dims=True,
                          code=130)
        # shape, no var
        self.assertRaises(ValueError, ncinfo, ncfile, shape=True)
        # shape, var does not exist
        self.assertRaises(ValueError, ncinfo, ncfile, shape=True,
                          var='is3')
        # shape, code does not exist
        self.assertRaises(ValueError, ncinfo, ncfile, shape=True,
                          code=130)
        # attributes, var does not exist
        self.assertRaises(ValueError, ncinfo, ncfile, attributes=True,
                          var='is3')
        # attributes, code does not exist
        self.assertRaises(ValueError, ncinfo, ncfile, attributes=True,
                          code=130)


class TestReadnetcdf(unittest.TestCase):
    """
    Tests for readnetcdf.py
    """

    def test_readnetcdf(self):
        import numpy as np
        import netCDF4 as nc
        from pyjams import readnetcdf

        ncfile = 'tests/test_readnetcdf.nc'
        ncfile1 = 'tests/test_readnetcdf1.nc'

        # var
        fout = readnetcdf(ncfile, var='is1')
        fsoll = np.full((2, 4), 1.)
        assert isinstance(fout, np.ndarray)
        self.assertEqual(_flatten(fout), _flatten(fsoll))

        # code
        fout = readnetcdf(ncfile, code=129)
        fsoll = np.full((2, 4), 2.)
        assert isinstance(fout, np.ndarray)
        self.assertEqual(_flatten(fout), _flatten(fsoll))

        # squeeze
        fout = readnetcdf(ncfile, var='is1', squeeze=True)
        fsoll = np.full((2, 4), 1.)
        assert isinstance(fout, np.ndarray)
        self.assertEqual(_flatten(fout), _flatten(fsoll))

        # pointer
        fh, var = readnetcdf(ncfile1, var='is1', pointer=True)
        fsoll = np.full((2, 4), 1.)
        assert isinstance(fh, nc.Dataset)
        assert isinstance(var, nc.Variable)
        self.assertEqual(var.shape, (2, 4))
        fout = var[:]
        self.assertEqual(_flatten(fout), _flatten(fsoll))
        fh.close()

        # overwrite
        fh, var = readnetcdf(ncfile1, var='is1', overwrite=True)
        fsoll = np.full((2, 4), 2.)
        assert isinstance(fh, nc.Dataset)
        assert isinstance(var, nc.Variable)
        self.assertEqual(var.shape, (2, 4))
        var[:] *= 2.
        fh.close()
        fout = readnetcdf(ncfile1, var='is1')
        self.assertEqual(_flatten(fout), _flatten(fsoll))
        fh, var = readnetcdf(ncfile1, var='is1', overwrite=True)
        var[:] *= 0.5
        fh.close()

        # errors
        # no var or code
        self.assertRaises(ValueError, readnetcdf, ncfile)
        # var does not exist
        self.assertRaises(ValueError, readnetcdf, ncfile, 'is3')
        # code does not exist
        self.assertRaises(ValueError, readnetcdf, ncfile, code=130)
        # #vars > 1 with overwrite
        self.assertRaises(ValueError, readnetcdf, ncfile, 'is1', overwrite=True)

    def test_ncread(self):
        import os
        import numpy as np
        import netCDF4 as nc
        from pyjams import ncread

        ncfile = 'tests/test_readnetcdf.nc'
        ncfile1 = 'tests/test_readnetcdf1.nc'

        # var
        fout = ncread(ncfile, var='is1')
        fsoll = np.full((2, 4), 1.)
        assert isinstance(fout, np.ndarray)
        self.assertEqual(_flatten(fout), _flatten(fsoll))

        # code
        fout = ncread(ncfile, code=129)
        fsoll = np.full((2, 4), 2.)
        assert isinstance(fout, np.ndarray)
        self.assertEqual(_flatten(fout), _flatten(fsoll))

        # squeeze
        fout = ncread(ncfile, var='is1', squeeze=True)
        fsoll = np.full((2, 4), 1.)
        assert isinstance(fout, np.ndarray)
        self.assertEqual(_flatten(fout), _flatten(fsoll))

        # pointer
        fh, var = ncread(ncfile1, var='is1', pointer=True)
        fsoll = np.full((2, 4), 1.)
        assert isinstance(fh, nc.Dataset)
        assert isinstance(var, nc.Variable)
        self.assertEqual(var.shape, (2, 4))
        fout = var[:]
        self.assertEqual(_flatten(fout), _flatten(fsoll))
        fh.close()

        # overwrite
        fh, var = ncread(ncfile1, var='is1', overwrite=True)
        fsoll = np.full((2, 4), 2.)
        assert isinstance(fh, nc.Dataset)
        assert isinstance(var, nc.Variable)
        self.assertEqual(var.shape, (2, 4))
        var[:] *= 2.
        fh.close()
        fout = ncread(ncfile1, var='is1')
        self.assertEqual(_flatten(fout), _flatten(fsoll))
        fh, var = ncread(ncfile1, var='is1', overwrite=True)
        var[:] *= 0.5
        fh.close()

        # errors
        # no var or code
        self.assertRaises(ValueError, ncread, ncfile)
        # var does not exist
        self.assertRaises(ValueError, ncread, ncfile, 'is3')
        # code does not exist
        self.assertRaises(ValueError, ncread, ncfile, code=130)
        # #vars > 1 with overwrite
        self.assertRaises(ValueError, ncread, ncfile, 'is1', overwrite=True)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
"""
This is the unittest for datetime module.

python -m unittest -v tests/test_datetime.py
python -m pytest --cov=pyjams --cov-report term-missing -v tests/test_datetime.py

"""
import unittest


def _flatten(itr):
    import numpy as np
    fitr = np.array(itr).flatten()
    if len(fitr) == 0:
        return list(fitr)
    else:
        if isinstance(fitr[0], str):
            return [ i for i in fitr ]
        else:
            return [ i if np.isfinite(i) else np.finfo(float).max
                     for i in fitr ]


class TestDatetime(unittest.TestCase):
    """
    Tests for datetime.py
    """

    def test_date2num2date(self):
        import random
        import numpy as np
        from pyjams import date2date
        from pyjams import date2num, num2date, date2dec, dec2date

        _excelcalendars = ['excel', 'excel1900', 'excel1904']
        _decimalcalendars = ['decimal', 'decimal360', 'decimal365', 'decimal366']
        _noncfcalendars = _excelcalendars + _decimalcalendars
        _cfcalendars = ['standard', 'gregorian', 'proleptic_gregorian',
                        'noleap', 'julian', 'all_leap', '365_day', '366_day',
                        '360_day']
        _idealized_cfcalendars = ['all_leap', 'noleap', '366_day', '365_day',
                                  '360_day']

        idates = ['2000-01-05 12:30:15', '1810-04-24 16:15:10',
                  '1630-07-15 10:20:40', '1510-09-20 14:35:50',
                  '1271-03-18 19:41:34', '0619-08-27 11:08:37',
                  '0001-01-01 12:00:00']
        iformat = '%Y-%m-%d %H:%M:%S'

        # Back and forth for _noncfcalendars

        # precision problem for some calendars of _cfcalendars, i.e.
        #     'noleap', 'all_leap', '365_day', '366_day', '360_day'
        calendars = _noncfcalendars
        units = ['', 'day as %Y%m%d.%f', 'month as %Y%m.%f', 'year as %Y.%f']
        formats = ['%Y-%m-%d %H:%M:%S', '%d.%m.%Y %H:%M:%S',
                   '%d/%m/%Y %H:%M:%S',
                   '%Y-%m-%dT%H:%M:%S', '%d.%m.%YT%H:%M:%S',
                   '%d/%m/%YT%H:%M:%S',
                   '%Y%m%d%H%M%S']
        only_cftimes = [True, False]
        only_pythons = [True, False]
        has_year_zeros = [None, True, False]
        itypes = [list, tuple, np.array]
        ttypes = [list, tuple, np.ndarray]
        f2dates = [num2date, dec2date]
        f2nums  = [date2num, date2dec]
        for calendar in calendars:
            for unit in units:
                for format in formats:
                    for only_cftime in only_cftimes:
                        for only_python in only_pythons:
                            for has_year_zero in has_year_zeros:
                                for itype in range(len(itypes)):
                                    # print(calendar, unit, format, only_cftime,
                                    #       only_python, has_year_zero,
                                    #       itypes[itype])
                                    indates = itypes[itype](idates)
                                    jdates = date2date(indates, format=format)
                                    f2num = f2nums[random.randint(0, 1)]
                                    idec = f2num(
                                        jdates, units=unit, calendar=calendar,
                                        has_year_zero=has_year_zero,
                                        format=format)
                                    f2date = f2dates[random.randint(0, 1)]
                                    odates = f2date(
                                        idec, units=unit, calendar=calendar,
                                        only_use_cftime_datetimes=only_cftime,
                                        only_use_python_datetimes=only_python,
                                        has_year_zero=has_year_zero,
                                        format=iformat, return_arrays=False)
                                    assert isinstance(jdates, ttypes[itype])
                                    assert isinstance(idec, ttypes[itype])
                                    assert isinstance(odates, ttypes[itype])
                                    self.assertEqual(_flatten(odates),
                                                     _flatten(indates))

        # Back and forth for _cfcalendars

        # precision problem for some calendars of _cfcalendars, i.e.
        #     'noleap', 'all_leap', '365_day', '366_day', '360_day'
        # calendars = _cfcalendars
        calendars = ['standard', 'gregorian', 'proleptic_gregorian', 'julian']
        units = ['']
        formats = ['%Y-%m-%d %H:%M:%S', '%d.%m.%Y %H:%M:%S',
                   '%d/%m/%Y %H:%M:%S',
                   '%Y-%m-%dT%H:%M:%S', '%d.%m.%YT%H:%M:%S',
                   '%d/%m/%YT%H:%M:%S',
                   '%Y%m%d%H%M%S']
        only_cftimes = [True, False]
        only_pythons = [False]  # [True, False]
        has_year_zeros = [None, True, False]
        itypes = [list, tuple, np.array]
        ttypes = [list, tuple, np.ndarray]
        f2dates = [num2date, dec2date]
        f2nums  = [date2num, date2dec]
        for calendar in calendars:
            for unit in units:
                for format in formats:
                    for only_cftime in only_cftimes:
                        for only_python in only_pythons:
                            for has_year_zero in has_year_zeros:
                                for itype in range(len(itypes)):
                                    # print(calendar, unit, format, only_cftime,
                                    #       only_python, has_year_zero,
                                    #       itypes[itype])
                                    indates = itypes[itype](idates)
                                    jdates = date2date(indates, format=format)
                                    f2num = f2nums[random.randint(0, 1)]
                                    idec = f2num(
                                        jdates, units=unit, calendar=calendar,
                                        has_year_zero=has_year_zero,
                                        format=format)
                                    f2date = f2dates[random.randint(0, 1)]
                                    odates = f2date(
                                        idec, units=unit, calendar=calendar,
                                        only_use_cftime_datetimes=only_cftime,
                                        only_use_python_datetimes=only_python,
                                        has_year_zero=has_year_zero,
                                        format=iformat, return_arrays=False)
                                    assert isinstance(jdates, ttypes[itype])
                                    assert isinstance(idec, ttypes[itype])
                                    assert isinstance(odates, ttypes[itype])
                                    self.assertEqual(_flatten(odates),
                                                     _flatten(indates))

        # edate = ['2014-11-12 12:00', '01.03.2015T17:56:00',
        #          '12/01/1990', '04.05.1786']

        # # errors
        # # wrong date delimiter
        # edate = ['2014\11\12 12:00', '01.03.2015 17:56:00',
        #          '1990-12-01 01', '04.05.1786']
        # self.assertRaises(ValueError, date2date, edate)
        # # milliseconds
        # edate = ['2014-11-12 12:00', '01.03.2015 17:56:00:00',
        #          '1990-12-01 01', '04.05.1786']
        # self.assertRaises(ValueError, date2date, edate)
        # # 3-digit year
        # edate = ['1-11-12 12:00', '1.3.15 17:56:00', '190-12-1']
        # self.assertRaises(ValueError, date2date, edate)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
"""
This is the unittest for the Logistic Functions module.

python -m unittest -v tests/test_logistic_function.py
python -m pytest --cov=pyjams --cov-report term-missing -v tests/test_logistic_function.py

"""
from __future__ import division, absolute_import, print_function
import unittest


class TestLogisticFunction(unittest.TestCase):
    """
    Tests for functions/logistic_function.py

    """

    def test_logistic_function(self):
        import numpy as np
        from pyjams.functions import logistic, logistic_offset
        from pyjams.functions import logistic2_offset
        from pyjams.functions import dlogistic, dlogistic_offset
        from pyjams.functions import dlogistic2_offset
        from pyjams.functions import d2logistic, d2logistic_offset
        from pyjams.functions import d2logistic2_offset
        from pyjams.functions import logistic_p, logistic_offset_p
        from pyjams.functions import logistic2_offset_p
        from pyjams.functions import dlogistic_p, dlogistic_offset_p
        from pyjams.functions import dlogistic2_offset_p
        from pyjams.functions import d2logistic_p, d2logistic_offset_p
        from pyjams.functions import d2logistic2_offset_p

        self.assertEqual(logistic(1.,  1., 0., 2.), 0.5)
        self.assertEqual(logistic(1.,  1., 2., 1.), 0.5)
        self.assertEqual(logistic(2.,  1., 1., 1.), 1./(1.+np.exp(-1.)))
        self.assertEqual(logistic_offset(1.,  1., 0., 2., 1.), 1.5)
        self.assertEqual(logistic_offset(1.,  1., 2., 1., 1.), 1.5)
        self.assertEqual(logistic_offset(2.,  1., 1., 1., 1.),
                         1./(1.+np.exp(-1.)) + 1.)
        self.assertEqual(logistic2_offset(1.,  1., 2., 1.,  2., 2., 1.,  1.),
                         0.5)
        self.assertEqual(dlogistic(1.,  1., 2., 1.), 0.5)
        self.assertEqual(dlogistic_offset(1.,  1., 2., 1., 1.), 0.5)
        self.assertEqual(dlogistic2_offset(1.,  1., 2., 1.,  2., 2., 1.,  1.),
                         -0.5)
        self.assertEqual(np.around(d2logistic(1., 1., 2., 2.), 4), 0.3199)
        self.assertEqual(np.around(d2logistic_offset(1., 1., 2., 2., 1.), 4),
                         0.3199)
        self.assertEqual(np.around(d2logistic2_offset(1., 1., 2., 2.,
                                                      2., 2., 2., 1.), 4),
                         -0.3199)

        self.assertEqual(logistic_p(1.,  [1., 0., 2.]), 0.5)
        self.assertEqual(logistic_p(1.,  [1., 2., 1.]), 0.5)
        self.assertEqual(logistic_p(2.,  [1., 1., 1.]), 1./(1.+np.exp(-1.)))
        self.assertEqual(logistic_offset_p(1.,  [1., 0., 2., 1.]), 1.5)
        self.assertEqual(logistic_offset_p(1.,  [1., 2., 1., 1.]), 1.5)
        self.assertEqual(logistic_offset_p(2.,  [1., 1., 1., 1.]),
                         1./(1.+np.exp(-1.)) + 1.)
        self.assertEqual(logistic2_offset_p(1.,
                                            [1., 2., 1.,  2., 2., 1.,  1.]),
                         0.5)
        self.assertEqual(dlogistic_p(1.,  [1., 2., 1.]), 0.5)
        self.assertEqual(dlogistic_offset_p(1.,  [1., 2., 1., 1.]), 0.5)
        self.assertEqual(dlogistic2_offset_p(1.,
                                             [1., 2., 1.,  2., 2., 1.,  1.]),
                         -0.5)
        self.assertEqual(np.around(d2logistic_p(1., [1., 2., 2.]), 4), 0.3199)
        self.assertEqual(np.around(d2logistic_offset_p(1.,
                                                       [1., 2., 2., 1.]), 4),
                         0.3199)
        self.assertEqual(np.around(d2logistic2_offset_p(1., [1., 2., 2.,  2.,
                                                             2., 2.,  1.]), 4),
                         -0.3199)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
"""
This is the unittest for the Sensitivity Analysis Test Functions module.

python -m unittest -v tests/test_sa_test_functions.py
python -m pytest --cov=pyjams --cov-report term-missing -v tests/test_sa_test_functions.py

"""
from __future__ import division, absolute_import, print_function
import unittest


class TestSATestFunctions(unittest.TestCase):
    """
    Tests for functions/sa_test_functions.py

    """
    def setUp(self):
        import numpy as np
        # seed for reproducible results
        seed = 1234
        np.random.seed(seed=seed)

    def test_sa_test_functions(self):
        import numpy as np
        from pyjams.functions import B, g, G, Gstar, K, bratley
        from pyjams.functions import oakley_ohagan, ishigami_homma
        from pyjams.functions import linear, product, ratio
        from pyjams.functions import ishigami_homma_easy, fmorris, morris

        # scalar
        self.assertEqual(B(np.arange(10)), 80)
        self.assertEqual(g(np.ones(5), np.zeros(5)), 32.0)
        self.assertEqual(G(np.ones(5), np.zeros(5)), 32.0)
        self.assertEqual(Gstar(np.ones(5), np.zeros(5), np.ones(5),
                               np.zeros(5)), 1.0)
        self.assertEqual(Gstar(np.ones(5), [0., 0., 0., 0., 0.],
                               np.ones(5), np.zeros(5)), 1.0)
        self.assertEqual(K(np.arange(5)+1.), -101.0)
        self.assertEqual(bratley(np.arange(5)+1.), -101.0)
        self.assertEqual(oakley_ohagan(np.zeros(15)), 15.75)
        self.assertEqual(ishigami_homma([np.pi/2., np.pi/2., 1.], 1., 1.), 3.0)
        self.assertEqual(linear(np.ones(1), 1., 1.), 2.0)
        self.assertEqual(product(np.arange(2)+1.), 2.0)
        self.assertEqual(ratio(np.arange(2)+1.), 0.5)
        self.assertEqual(ishigami_homma_easy([np.pi/2., 1.]), 2.0)

        # vector
        self.assertEqual(list(B(np.arange(12).reshape(6, 2))), [56, 89])
        self.assertEqual(list(g(np.ones((5, 2)), np.zeros(5))), [32.0, 32.0])
        self.assertEqual(list(G(np.ones((5, 2)), np.zeros(5))), [32.0, 32.0])
        self.assertEqual(list(Gstar(np.ones((5, 2)), np.zeros(5), np.ones(5),
                                    np.zeros(5))), [1.0, 1.0])
        self.assertEqual(list(K(np.arange(8).reshape((4, 2))+1.)), [92., 342.])
        self.assertEqual(list(bratley(np.arange(8).reshape((4, 2))+1.)),
                         [92., 342.])
        self.assertEqual(list(oakley_ohagan(np.zeros((15, 2)))),
                         [15.75, 15.75])
        self.assertEqual(list(ishigami_homma([[np.pi/2., np.pi/2.],
                                              [np.pi/2., np.pi/2.],
                                              [1., 1.]], 1., 1.)),
                         [3.0, 3.0])
        self.assertEqual(list(linear(np.ones((1, 2)), 1., 1.)), [2.0, 2.0])
        self.assertEqual(list(product(np.arange(4).reshape((2, 2))+1.)),
                         [3.0, 8.0])
        self.assertEqual(list(
            ratio(np.arange(2).repeat(2).reshape((2, 2))+1.)), [0.5, 0.5])
        self.assertEqual(list(ishigami_homma_easy([[np.pi/2., np.pi/2.],
                                                   [1., 1.]])), [2.0, 2.0])

        # Morris
        npars = 20
        x0 = np.ones(npars)*0.5
        lb = np.zeros(npars)
        ub = np.ones(npars)
        beta0                 = 0.
        beta1                 = np.random.standard_normal(npars)
        beta1[:10]            = 20.
        beta2                 = np.random.standard_normal((npars, npars))
        beta2[:6, :6]         = -15.
        beta3                 = np.zeros((npars, npars, npars))
        beta3[:5, :5, :5]     = -10.
        beta4                 = np.zeros((npars, npars, npars, npars))
        beta4[:4, :4, :4, :4] = 5.

        mm = fmorris(np.linspace(0, 2*(npars-1), npars)/float(2*npars-1),
                     beta0, beta1, beta2, beta3, beta4)
        self.assertEqual(np.around(mm, 3), -82.711)
        mm = fmorris(np.arange(2*npars, dtype=float).reshape((npars, 2)) /
                     float(2*npars-1),
                     beta0, beta1, beta2, beta3, beta4)
        self.assertEqual(list(np.around(mm, 3)), [-82.711, -60.589])

        mm = morris(np.linspace(0, 2*(npars-1), npars)/float(2*npars-1),
                    beta0, beta1, beta2, beta3, beta4)
        self.assertEqual(np.around(mm, 3), -82.711)
        mm = morris(np.arange(2*npars, dtype=float).reshape((npars, 2)) /
                    float(2*npars-1),
                    beta0, beta1, beta2, beta3, beta4)
        self.assertEqual(list(np.around(mm, 3)), [-82.711, -60.589])


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
"""
Get, show, print color palettes

This module was written by Matthias Cuntz while at Institut National de
Recherche pour l'Agriculture, l'Alimentation et l'Environnement (INRAE), Nancy,
France.

:copyright: Copyright 2021-2022 Matthias Cuntz, see AUTHORS.rst for details.
:license: MIT License, see LICENSE for details.

.. moduleauthor:: Matthias Cuntz

The following functions are provided:

.. autosummary::
   get_color
   print_colors
   get_cmap
   print_palettes
   show_palettes

History
    * Written Nov 2021, Matthias Cuntz
    * Change order of color maps, Dec 2021, Matthias Cuntz
    * More consistent docstrings, Jan 2022, Matthias Cuntz
    * Added 'order' keyword, Feb 2022, Matthias Cuntz
    * 'get_color', Mar 2022, Matthias Cuntz
    * 'cname' can be list of names in 'get_color', Apr 2022, Matthias Cuntz
    * Register ufz colors only once, Apr 2022, Matthias Cuntz
    * Added print_colors, Apr 2022, Matthias Cuntz

"""

__all__ = ['get_color', 'print_colors',
           'get_cmap', 'print_palettes', 'show_palettes']


def _rgb2rgb(col):
    """
    Transform RGB tuple with values 0-255 to tuple with values 0-1

    Parameters
    ----------
    col : iterable
        RGB tuple with values from 0-255

    Returns
    -------
    tuple
        RGB tuple with values from 0-1

    Examples
    --------
    col = (0, 126, 255)
    col = _rgb2rgb(col)

    """
    return tuple([ i / 255. for i in col ])


def _to_grey(col):
    """
    Transform RGB tuple to grey values

    Parameters
    ----------
    col : iterable
        RGB tuple

    Returns
    -------
    tuple
        RGB tuple with corresponding grey values

    Examples
    --------
    col = (0, 0.5, 1)
    col = _to_grey(col)

    """
    isgrey = 0.2125 * col[0] + 0.7154 * col[1] + 0.072 * col[2]
    return (isgrey, isgrey, isgrey)


def get_color(cname=''):
    """
    Register new named colors with Matplotlib and return named color(s)

    Registers the named colors given in ufz_palettes.
    Optionally returns the value(s) of the named color(s),
    which can be any color name known to Matplotlib.

    Parameters
    ----------
    cname : str or iterable of str, optional
        Colour name(s)

    Examples
    --------
    .. code-block:: python

       col1 = get_color('ufz:blue')
       col2 = get_color('xkcd:blue')
       cols = get_color(['blue', 'red'])

    """
    from collections.abc import Iterable
    import matplotlib.pyplot as plt
    import pyjams.color

    mapping = plt.cm.colors.get_named_colors_mapping
    # register ufz colors if not yet done
    if 'ufz:blue' not in mapping().keys():
        ufz_colors = [ i for i in dir(pyjams.color)
                       if i.startswith('ufz_')
                       and not i.endswith('_palettes') ]
        # ufz has only 1 list with colors
        ufz_colors = eval('pyjams.color.' + ufz_colors[0])
        mapping().update(ufz_colors)

    if cname:
        if isinstance(cname, Iterable) and (not isinstance(cname, str)):
            out = [ mapping()[i] for i in cname ]
            return out
        else:
            return mapping()[cname]


def print_colors(collection=''):
    """
    Print the known named colors

    Parameters
    ----------
    collection : str or list of strings, optional
        Name(s) of color collection(s).
        Known collections are 'base', 'tableau', 'ufz', 'css',
        and 'xkcd'.

    Returns
    -------
    None
        Prints list of known named colors to console.

    Examples
    --------
    .. code-block:: python

       print_colors()

    """
    import matplotlib.pyplot as plt
    import pyjams.color

    if collection:
        if isinstance(collection, str):
            collections = [collection.lower()]
        else:
            collections = [ i.lower for i in collection ]
    else:
        collections = ['base', 'tableau', 'ufz', 'css', 'xkcd']

    # register ufz colors if needed
    get_color()
    mapping = plt.cm.colors.get_named_colors_mapping
    colors = mapping().keys()

    if 'base' in collections:
        print('base')
        ll = [ cc for cc in colors if len(cc) == 1 ]
        print('       ', list(ll))

    if 'tableau' in collections:
        print('tableau')
        ll = [ cc for cc in colors if cc.startswith('tab:') ]
        print('       ', list(ll))

    if 'ufz' in collections:
        print('ufz')
        ll = [ cc for cc in colors if cc.startswith('ufz:') ]
        print('       ', list(ll))

    if 'css' in collections:
        print('css')
        ll = [ cc for cc in colors if (':' not in cc) and (len(cc) > 1) ]
        print('       ', list(ll))

    if 'xkcd' in collections:
        print('xkcd')
        ll = [ cc for cc in colors if cc.startswith('xkcd:') ]
        print('       ', list(ll))


def get_cmap(palette, ncol=0, offset=0, upper=1, as_cmap=False,
             reverse=False, grey=False, order=None):
    """
    Get colors of defined palettes or continuous colormaps

    Parameters
    ----------
    palette : str
        Name of color palette or continuous colormap
    ncol : int, optional
        Number of desired colors.
        If 0, all colors defined by the specific palette will be returned.
        256 colors will be chosen for continuous colormaps.
        If > 0, existing color palettes will be subsampled to *ncol* colors.
        *ncol* colors will be produced from continuous colormaps.
    offset : float (0-1), optional
        Bottom fraction to exclude for subsample or continuous colormaps
    upper : float (0-1), optional
        Upper most fraction to include for subsample or continuous colormaps
    reverse : bool, optional
        Reverse colormap if True (default: False). This can also be achieved by
        adding '_r' to the palette name. Palettes that end with '_r' will not
        be reversed.
    grey : bool, optional
        Return grey equivalent colors if True (default: False).
    order : str, optional
        Order colors by 'hue', 'saturation', or 'value'.
        This is done before *reverse* and *grey*
        (default: order from color sources).

    Returns
    -------
    list or matplotlib.colors.Colormap
        List of RGB tuples or :class:`matplotlib.colors.Colormap`

    Notes
    -----
    `get_cmap` always returns a list of RGB tuples or a Matplotlib
    ListedColormap. You can use the *ncol*, *offset*, and *upper* keywords to
    subsample colormaps. To get smoothly-varying colormaps, you can use the
    method :meth:`matplotlib.colors.LinearSegmentedColormap.from_list`

       cols = get_cmap('sron_ylorbr')

       cmap = matplotlib.colors.LinearSegmentedColormap.from_list(cols)

    You can get the color tuples from a LinearSegmentedColormap like this

       colors = [ cmap(i) for i in range(cmap.N) ]

    Examples
    --------
    .. code-block:: python

       cols = get_cmap('mathematica_dark_rainbow_256', 15)

    """
    import matplotlib as mpl
    import matplotlib.pyplot as plt
    import pyjams.color
    from pyjams import argsort

    # _r at end of palette is same as reverse=True
    if palette.endswith('_r'):
        palette = palette[:-2]
        reverse = True

    pyjams_collections = [ i for i in dir(pyjams.color)
                           if i.startswith('pyjams_')
                           and not i.endswith('_palettes') ]
    sron_collections = [ i for i in dir(pyjams.color)
                         if i.startswith('sron_')
                         and not i.endswith('_palettes') ]
    sron2012_collections = [ i for i in dir(pyjams.color)
                             if i.startswith('sron2012_')
                             and not i.endswith('_palettes') ]
    mathematica_collections = [ i for i in dir(pyjams.color)
                                if i.startswith('mathematica_')
                                and not i.endswith('_palettes') ]
    oregon_collections = [ i for i in dir(pyjams.color)
                           if i.startswith('oregon_')
                           and not i.endswith('_palettes') ]
    ncl_collections = [ i for i in dir(pyjams.color)
                        if i.startswith('ncl_')
                        and not i.endswith('_palettes') ]
    brewer_collections = [ i for i in dir(pyjams.color)
                           if i.startswith('brewer_')
                           and not i.endswith('_palettes') ]

    found_palette = False
    nosubsample = False
    miss = None

    if not found_palette:
        for bb in pyjams_collections:
            dd = eval('pyjams.color.' + bb)
            if palette in dd:
                found_palette = True
                colors = dd[palette]

    if not found_palette:
        for bb in sron_collections:
            dd = eval('pyjams.color.' + bb)
            if palette in dd:
                found_palette = True
                if bb == 'sron_colors':
                    colors = [ mpl.colors.colorConverter.to_rgb(i)
                               for i in dd[palette] ]
                elif bb == 'sron_colormaps':
                    colors = [ mpl.colors.colorConverter.to_rgb(i)
                               for i in dd[palette][0] ]
                    miss = mpl.colors.colorConverter.to_rgb(dd[palette][1])
                elif bb == 'sron_functions':
                    nosubsample = True
                    if ncol == 0:
                        ncol1 = 23
                    else:
                        ncol1 = ncol
                    cols = dd[palette](ncol1)
                    colors = [ mpl.colors.colorConverter.to_rgb(i)
                               for i in cols[0] ]
                    miss = mpl.colors.colorConverter.to_rgb(cols[1])

    if not found_palette:
        for bb in sron2012_collections:
            dd = eval('pyjams.color.' + bb)
            if palette in dd:
                found_palette = True
                if bb == 'sron2012_colors':
                    colors = [ mpl.colors.colorConverter.to_rgb(i)
                               for i in dd[palette] ]
                elif bb == 'sron2012_functions':
                    nosubsample = True
                    colors = []
                    if ncol == 0:
                        ncol1 = 256
                    else:
                        ncol1 = ncol
                    for i in range(ncol1):
                        x = offset + float(i)/float(ncol1-1) * (upper-offset)
                        colors.append(tuple(dd[palette](x)))

    if not found_palette:
        for bb in mathematica_collections:
            dd = eval('pyjams.color.' + bb)
            if palette in dd:
                found_palette = True
                colors = dd[palette]

    if not found_palette:
        for bb in oregon_collections:
            dd = eval('pyjams.color.' + bb)
            if palette in dd:
                found_palette = True
                colors = [ _rgb2rgb(i) for i in dd[palette] ]

    if not found_palette:
        for bb in ncl_collections:
            dd = eval('pyjams.color.' + bb)
            if palette in dd:
                found_palette = True
                colors = dd[palette]

    if not found_palette:
        # try:
        #     amplmaps = plt.colormaps()
        #     mplmaps = [ i for i in amplmaps if not i.endswith('_r') ]
        # except AttributeError:
        #     mplmaps = sorted(mpl.cm.datad.keys())
        amplmaps = plt.colormaps()
        mplmaps = [ i for i in amplmaps if not i.endswith('_r') ]
        if palette in mplmaps:
            found_palette = True
            cmap = mpl.cm.get_cmap(palette)
            try:
                # mpl.colors.ListedColormap
                colors = cmap.colors
            except AttributeError:
                # mpl.colors.LinearSegmentedColormap
                colors = [ cmap(i) for i in range(cmap.N) ]
        else:
            lmplmaps = [ i.lower() for i in mplmaps ]
            if palette.lower() in lmplmaps:
                found_palette = True
                ipalette = lmplmaps.index(palette.lower())
                cmap = mpl.cm.get_cmap(mplmaps[ipalette])
                try:
                    # mpl.colors.ListedColormap
                    colors = cmap.colors
                except AttributeError:
                    # mpl.colors.LinearSegmentedColormap
                    colors = [ cmap(i) for i in range(cmap.N) ]

    if not found_palette:
        for bb in brewer_collections:
            dd = eval('pyjams.color.' + bb)
            if palette in dd:
                found_palette = True
                colors = [ _rgb2rgb(i) for i in dd[palette] ]

    if not found_palette:
        raise ValueError(palette+' color palette not found.')

    if order is not None:
        hsv = ['hue', 'saturation', 'value']
        if order.lower() in hsv:
            oo = hsv.index(order.lower())
        else:
            raise ValueError('Sort order not known: ' + order)
        ii = argsort([ mpl.colors.rgb_to_hsv(cc)[oo]
                       for cc in colors ])
        colors = [ colors[i] for i in ii ]

    if reverse:
        colors = colors[::-1]

    if grey:
        colors = [ _to_grey(i) for i in colors ]

    # subsample
    if (ncol > 0) and not nosubsample:
        ncolors = len(colors)
        ocolors = ['']*ncol
        for i in range(ncol):
            x = offset + float(i)/float(ncol-1) * (upper-offset)  # [0-1]
            iicol = round(x * (ncolors-1))  # [0-ncolor-1]
            ocolors[i] = colors[iicol]
        colors = ocolors

    if as_cmap:
        colors = mpl.colors.ListedColormap(colors)
        if mpl.__version__ > '3.4.0':
            if miss is not None:
                colors.set_bad(miss)

    return colors


def print_palettes(collection=''):
    """
    Print the known color palettes and continuous colormaps

    Parameters
    ----------
    collection : str or list of strings, optional
        Name(s) of color palette collection(s).
        Known collections are 'pyjams', 'sron', 'sron2012', 'mathematica',
        'oregon', 'ncl', 'matplotlib', and 'brewer'.

    Returns
    -------
    None
        Prints list of known color palettes and continuous colormaps to
        console.

    Examples
    --------
    .. code-block:: python

       print_palettes()

    """
    import matplotlib as mpl
    import matplotlib.pyplot as plt
    import pyjams.color

    pyjams_collections = [ i for i in dir(pyjams.color)
                           if i.startswith('pyjams_')
                           and not i.endswith('_palettes') ]
    sron_collections = [ i for i in dir(pyjams.color)
                         if i.startswith('sron_')
                         and not i.endswith('_palettes') ]
    sron2012_collections = [ i for i in dir(pyjams.color)
                             if i.startswith('sron2012_')
                             and not i.endswith('_palettes') ]
    mathematica_collections = [ i for i in dir(pyjams.color)
                                if i.startswith('mathematica_')
                                and not i.endswith('_palettes') ]
    oregon_collections = [ i for i in dir(pyjams.color)
                           if i.startswith('oregon_')
                           and not i.endswith('_palettes') ]
    ncl_collections = [ i for i in dir(pyjams.color)
                        if i.startswith('ncl_')
                        and not i.endswith('_palettes') ]
    brewer_collections = [ i for i in dir(pyjams.color)
                           if i.startswith('brewer_')
                           and not i.endswith('_palettes') ]

    if collection:
        if isinstance(collection, str):
            collections = [collection.lower()]
        else:
            collections = [ i.lower for i in collection ]
    else:
        collections = ['pyjams', 'sron', 'sron2012', 'mathematica',
                       'oregon', 'ncl', 'matplotlib', 'brewer']

    if 'pyjams' in collections:
        print('pyjams')
        for cc in pyjams_collections:
            print('   ', cc)
            ll = eval('pyjams.color.' + cc + '.keys()')
            print('       ', list(ll))

    if 'sron' in collections:
        print('sron')
        for cc in sron_collections:
            print('   ', cc)
            ll = eval('pyjams.color.' + cc + '.keys()')
            print('       ', list(ll))

    if 'sron2012' in collections:
        print('sron2012')
        for cc in sron2012_collections:
            print('   ', cc)
            ll = eval('pyjams.color.' + cc + '.keys()')
            print('       ', list(ll))

    if 'mathematica' in collections:
        print('mathematica')
        for cc in mathematica_collections:
            print('   ', cc)
            ll = eval('pyjams.color.' + cc + '.keys()')
            print('       ', list(ll))

    if 'oregon' in collections:
        print('oregon')
        for cc in oregon_collections:
            print('   ', cc)
            ll = eval('pyjams.color.' + cc + '.keys()')
            print('       ', list(ll))

    if 'ncl' in collections:
        print('ncl')
        for cc in ncl_collections:
            print('   ', cc)
            ll = eval('pyjams.color.' + cc + '.keys()')
            print('       ', list(ll))

    if 'matplotlib' in collections:
        print('matplotlib')
        # try:
        #     acmaps = plt.colormaps()
        #     cmaps  = [ i for i in acmaps if not i.endswith('_r') ]
        #     cmaps.sort()
        # except AttributeError:
        #     cmaps = sorted(mpl.colorbar.cm.datad.keys())
        acmaps = plt.colormaps()
        cmaps  = [ i for i in acmaps if not i.endswith('_r') ]
        cmaps.sort()
        print('   ', cmaps)

    if 'brewer' in collections:
        print('brewer')
        for cc in brewer_collections:
            print('   ', cc)
            ll = eval('pyjams.color.' + cc + '.keys()')
            print('       ', list(ll))


def _newfig(ifig, ititle):  # pragma: no cover
    """ Helper function for show_palettes """
    import matplotlib.pyplot as plt
    fig = plt.figure(ifig)
    fig.suptitle(ititle)
    plt.subplots_adjust(left=0.3, bottom=0.1,
                        right=0.95, top=0.95,
                        wspace=0.05, hspace=0.5)
    return fig


def _newsubplot(nrow, ncol, iplot, iname, ncolors=0):  # pragma: no cover
    """ Helper function for show_palettes """
    import numpy as np
    import matplotlib.pyplot as plt
    ax = plt.subplot(nrow, ncol, iplot)
    ax.axis('off')
    cmap = get_cmap(iname, ncolors, as_cmap=True)
    # ax.pcolormesh(np.outer(np.ones(cmap.N), np.arange(cmap.N)),
    #               cmap=cmap)
    ax.imshow(np.outer(np.ones(cmap.N), np.arange(cmap.N)),
              aspect='auto', cmap=cmap, origin="lower")
    xmin, xmax = ax.get_xlim()
    ymin, ymax = ax.get_ylim()
    dx = -0.01
    dy = 0.5
    if ncolors:
        iiname = iname + '(' + str(ncolors) + ')'
    else:
        iiname = iname
    ax.text(xmin+dx*(xmax-xmin), ymin+dy*(ymax-ymin), iiname,
            ha='right', va='center')
    # return ax


def _savefig(fig, ifig, outtype, outfile, pdf_pages):  # pragma: no cover
    """ Helper function for show_palettes """
    import matplotlib.pyplot as plt
    if (outtype == 'pdf'):
        pdf_pages.savefig(fig)
        plt.close(fig)
    elif not (outtype == 'X'):
        ofile = (outfile[:outfile.rfind('.')] + '_' + '{:04d}'.format(ifig)
                 + '.' + outtype)
        fig.savefig(ofile)
        plt.close(fig)
    else:
        pass


def show_palettes(outfile='', collection=''):  # pragma: no cover
    """
    Show the known color palettes and continuous colormaps

    Parameters
    ----------
    outfile : str, optional
        Output file name. Output type will be determined from file suffix.
    collection : str or list of strings, optional
        Name(s) of color palette collection(s).
        All palettes will be shown if collection is empty or 'all'.
        Known collections are: 'pyjams', 'sron', 'sron2012', 'mathematica',
        'oregon', 'ncl', 'matplotlib', and 'brewer'.

    Returns
    -------
    None
        Plots known color palettes and continuous colormaps in file or on
        plotting window.

    Examples
    --------
    .. code-block:: python

       show_palettes(outfile='pyjams_cmaps.pdf',
                     collection=['mathematica', 'matplotlib'])

    """
    import matplotlib as mpl
    import matplotlib.pyplot as plt
    import pyjams.color
    # outtype
    if '.' in outfile:
        outtype = outfile[outfile.rfind('.')+1:]
        if outtype == 'pdf':
            mpl.use('PDF')  # set directly after import matplotlib
            from matplotlib.backends.backend_pdf import PdfPages
            textsize = 10
        else:
            mpl.use('Agg')  # set directly after import matplotlib
            textsize = 10
    else:
        outtype = 'X'
        textsize = 8
    import matplotlib.pyplot as plt

    # which collections to include
    collections = ['pyjams', 'sron', 'sron2012', 'mathematica', 'oregon',
                   'ncl', 'matplotlib', 'brewer']
    if collection:
        if isinstance(collection, str):
            if collection.lower() != 'all':
                collections = [collection.lower()]
        else:
            collections = [ i.lower() for i in collection ]

    # plot setup
    if outfile:
        mpl.rc('figure', figsize=(8.27, 11.69))  # a4 portrait
        if not (outtype == 'pdf'):
            mpl.rc('savefig', dpi=300, format=outtype)
    else:
        mpl.rc('figure', figsize=(8.27*0.75, 11.69*0.75))
    figsize = mpl.rcParams['figure.figsize']
    mpl.rc('font', size=textsize)
    nrow = 35

    # get collections
    pyjams_collections = [ i for i in dir(pyjams.color)
                           if i.startswith('pyjams_')
                           and not i.endswith('_palettes') ]
    sron_collections = [ i for i in dir(pyjams.color)
                         if i.startswith('sron_')
                         and not i.endswith('_palettes') ]
    sron2012_collections = [ i for i in dir(pyjams.color)
                             if i.startswith('sron2012_')
                             and not i.endswith('_palettes') ]
    mathematica_collections = [ i for i in dir(pyjams.color)
                                if i.startswith('mathematica_')
                                and not i.endswith('_palettes') ]
    oregon_collections = [ i for i in dir(pyjams.color)
                           if i.startswith('oregon_')
                           and not i.endswith('_palettes') ]
    ncl_collections = [ i for i in dir(pyjams.color)
                        if i.startswith('ncl_')
                        and not i.endswith('_palettes') ]
    brewer_collections = [ i for i in dir(pyjams.color)
                           if i.startswith('brewer_')
                           and not i.endswith('_palettes') ]

    all_collections = []
    if 'pyjams' in collections:
        for cc in pyjams_collections:
            all_collections.append(cc)
    if 'sron' in collections:
        for cc in sron_collections:
            all_collections.append(cc)
    if 'sron2012' in collections:
        for cc in sron2012_collections:
            all_collections.append(cc)
    if 'mathematica' in collections:
        for cc in mathematica_collections:
            all_collections.append(cc)
    if 'oregon' in collections:
        for cc in oregon_collections:
            all_collections.append(cc)
    if 'ncl' in collections:
        for cc in ncl_collections:
            all_collections.append(cc)
    if 'matplotlib' in collections:
        # try:
        #     acmaps = plt.colormaps()
        #     cmaps  = [ i for i in acmaps if not i.endswith('_r') ]
        #     cmaps.sort()
        # except AttributeError:
        #     cmaps = sorted(mpl.colorbar.cm.datad.keys())
        acmaps = plt.colormaps()
        cmaps  = [ i for i in acmaps if not i.endswith('_r') ]
        cmaps.sort()
        all_collections.extend(cmaps)
    if 'brewer' in collections:
        for cc in brewer_collections:
            all_collections.append(cc)

    # plotting
    ifig = 0
    if (outtype == 'pdf'):
        pdf_pages = PdfPages(outfile)
    else:
        pdf_pages = None

    ipanel = 0
    for cc in all_collections:
        try:
            icoll = eval('pyjams.color.' + cc)
            newcoll = True
            collname = cc
        except AttributeError:
            # matplotlib name
            icoll = [cc]
            newcoll = False
            collname = 'matplotlib'
        for iname in icoll:
            # colors = icoll[iname]
            if ipanel == 0:
                ifig += 1
                fig = _newfig(ifig, collname)
            ipanel += 1
            # ax = _newsubplot(nrow, 1, ipanel, iname)
            _newsubplot(nrow, 1, ipanel, iname)
            if collname == 'sron_functions':
                for ncolors in range(1, 24):
                    ipanel += 1
                    _newsubplot(nrow, 1, ipanel, iname, ncolors)
            if ipanel == nrow:
                _savefig(fig, ifig, outtype, outfile, pdf_pages)
                ipanel = 0
        if (ipanel != 0) and newcoll:
            _savefig(fig, ifig, outtype, outfile, pdf_pages)
            ipanel = 0
    if ipanel != 0:
        _savefig(fig, ifig, outtype, outfile, pdf_pages)

    if outtype == 'pdf':
        pdf_pages.close()
    elif outtype == 'X':
        plt.show()
    else:
        pass


if __name__ == '__main__':
    import doctest
    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)



#!/usr/bin/env python
"""
This is the unittest for the Optimisation Test Functions module.

python -m unittest -v tests/test_opti_test_functions.py
python -m pytest --cov=pyjams --cov-report term-missing -v tests/test_opti_test_functions.py

"""
from __future__ import division, absolute_import, print_function
import unittest


class TestOptiTestFunctions(unittest.TestCase):
    """
    Tests for functions/opti_test_functions.py

    """
    def setUp(self):
        import numpy as np
        # seed for reproducible results
        seed = 1234
        np.random.seed(seed=seed)

    def test_sa_test_functions(self):
        import numpy as np
        from pyjams.functions import ackley, griewank, goldstein_price
        from pyjams.functions import rastrigin, rosenbrock
        from pyjams.functions import six_hump_camelback

        # optima
        self.assertTrue(np.isclose(ackley(np.zeros(15)), 0.0))
        self.assertEqual(griewank(np.zeros(2)), 0.0)
        self.assertEqual(griewank(np.zeros(10)), 0.0)
        self.assertEqual(goldstein_price([0.0, -1.0]), 3.0)
        self.assertEqual(rastrigin(np.zeros(2)), -2.0)
        self.assertEqual(rosenbrock(np.ones(2)), 0.0)
        self.assertEqual(np.around(six_hump_camelback([-0.08983, 0.7126]), 4),
                         -1.0316)
        self.assertEqual(np.around(six_hump_camelback([0.08983, -0.7126]), 4),
                         -1.0316)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
"""
This is the unittest for esat module.

python -m unittest -v tests/test_esat.py
python -m pytest --cov=pyjams --cov-report term-missing -v tests/test_esat.py

"""
import unittest


class TestEsat(unittest.TestCase):
    """
    Tests for esat.py
    """

    def test_esat(self):
        import numpy as np
        from numpy.ma import masked
        from pyjams import esat
        import warnings

        T0 = 273.15

        # scalar
        T1 = T0 + 20.
        assert isinstance(esat(T1), float)
        assert np.around(esat(T1), 3) == 2335.847
        T1 = T0 - 20.
        assert isinstance(esat(T1), float)
        assert np.around(esat(T1), 3) == 103.074

        # list - different formulas with different word cases
        T = [20., -20.]
        T1 = [ tt + T0 for tt in T ]
        assert isinstance(esat(T1), list)
        self.assertEqual(
            list(np.around(esat(T1), 3)),
            [2335.847, 103.074])
        self.assertEqual(
            list(np.around(esat(T1, formula='GoffGratch'), 3)),
            [2335.847, 103.074])
        self.assertEqual(
            list(np.around(esat(T1, formula='MartiMauersberger'), 3)),
            [2335.847, 103.650])
        self.assertEqual(
            list(np.around(esat(T1, formula='MagnusTeten'), 3)),
            [2335.201, 102.771])
        self.assertEqual(
            list(np.around(esat(T1, formula='buck'), 3)),
            [2338.340, 103.286])
        self.assertEqual(
            list(np.around(esat(T1, formula='Buck_original'), 3)),
            [2337.282, 103.267])
        self.assertEqual(
            list(np.around(esat(T1, formula='wmo'), 3)),
            [2337.080, 103.153])
        self.assertEqual(
            list(np.around(esat(T1, formula='WEXLER'), 3)),
            [2323.254, 103.074])
        self.assertEqual(
            list(np.around(esat(T1, formula='Sonntag'), 3)),
            [2339.249, 103.249])
        self.assertEqual(
            list(np.around(esat(T1, formula='Bolton'), 3)),
            [2336.947, 103.074])
        self.assertEqual(
            list(np.around(esat(T1, formula='Fukuta'), 3)),
            [2335.847, 103.074])
        self.assertEqual(
            list(np.around(esat(T1, formula='HylandWexler'), 3)),
            [2338.804, 103.260])
        self.assertEqual(
            list(np.around(esat(T1, formula='IAPWS'), 3)),
            [2339.194, 103.074])
        self.assertEqual(
            list(np.around(esat(T1, formula='MurphyKoop'), 3)),
            [2339.399, 103.252])

        # tuple
        T1 = tuple(T1)
        assert isinstance(esat(T1), tuple)
        # ndarray
        T  = np.array(T)
        T1 = T + T0
        assert isinstance(esat(T1), np.ndarray)
        # masked_array
        T1 = np.ma.array(T + T0, mask=(T == 20.))
        es = esat(T1)
        assert isinstance(es, np.ndarray)
        assert isinstance(es, np.ma.MaskedArray)
        self.assertEqual(
            list(np.around(esat(T1), 3)),
            [masked, 103.074])

        # liquid
        T1 = [ tt + T0 for tt in T ]
        self.assertEqual(
            list(np.around(esat(T1, liquid=True), 3)),
            [2335.847, 125.292])
        self.assertEqual(
            list(np.around(esat(T1, formula='Fukuta', liquid=True), 3)),
            [2335.847, 125.079])

        # undef, scalar
        T1 = -9998.
        es = esat(T1, undef=-9998.)
        assert isinstance(es, float)
        self.assertEqual(es, -9998.)
        T1 = -9999.
        es = esat(T1)
        assert isinstance(es, float)
        self.assertEqual(es, -9999.)

        # undef, list
        T1 = [20. + T0, -9998.]
        es = esat(T1, undef=-9998.)
        assert isinstance(es, list)
        self.assertEqual(
            list(np.around(es, 3)),
            [2335.847, -9998.])
        T1 = [20. + T0, -9999.]
        es = esat(T1)
        assert isinstance(es, list)
        self.assertEqual(
            list(np.around(es, 3)),
            [2335.847, -9999.])

        # undef, tuple
        T1 = (20. + T0, -9998.)
        es = esat(T1, undef=-9998.)
        assert isinstance(es, tuple)
        self.assertEqual(
            list(np.around(es, 3)),
            [2335.847, -9998.])
        T1 = (20. + T0, -9999.)
        es = esat(T1)
        assert isinstance(es, tuple)
        self.assertEqual(
            list(np.around(es, 3)),
            [2335.847, -9999.])

        # undef, ndarray
        T1 = np.array([20. + T0, -9998.])
        es = esat(T1, undef=-9998.)
        assert isinstance(es, np.ndarray)
        self.assertEqual(
            list(np.around(es, 3)),
            [2335.847, -9998.])
        T1 = np.array([20. + T0, -9999.])
        es = esat(T1)
        assert isinstance(es, np.ndarray)
        self.assertEqual(
            list(np.around(es, 3)),
            [2335.847, -9999.])

        # undef, masked array
        T = np.array([20. + T0, -9998., -9999.])
        T1 = np.ma.array(T, mask=(T == -9999.))
        es = esat(T1, undef=-9998.)
        assert isinstance(es, np.ma.MaskedArray)
        self.assertEqual(
            list(np.around(es, 3)),
            [2335.847, masked, masked])
        T = np.array([20. + T0, -9998., -9999.])
        T1 = np.ma.array(T, mask=(T == -9998.))
        es = esat(T1)
        assert isinstance(es, np.ma.MaskedArray)
        self.assertEqual(
            list(np.around(es, 3)),
            [2335.847, masked, masked])

        # errors
        # T < 0
        T1 = T0 - T0
        self.assertRaises(AssertionError, esat, T1)
        # T > 373.15
        T1 = T0 + 101.
        self.assertRaises(AssertionError, esat, T1)
        # wrong formula
        T1 = T0 + 20.
        self.assertRaises(ValueError, esat, T1, formula='dummy')
        T1 = T0 - 20.
        self.assertRaises(ValueError, esat, T1, formula='dummy')

        # warning T < 100
        T1 = T0 - 200.
        with warnings.catch_warnings(record=True) as w:
            es = esat(T1)
            assert len(w) > 0


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
"""
This is the unittest for the Fit Functions module.

python -m unittest -v tests/test_fit_functions.py
python -m pytest --cov=pyjams --cov-report term-missing -v tests/test_fit_functions.py

"""
from __future__ import division, absolute_import, print_function
import unittest


class TestFitFunctions(unittest.TestCase):
    """
    Tests for functions/logistic_function.py
    Missing coverage:
        ca. 200ff: exception for import at different levels

    """

    def test_cost_abs_square(self):
        import numpy as np
        from pyjams.functions import logistic_p, cost_logistic, cost2_logistic
        from pyjams.functions import cost_abs, cost_square

        p = [1., 1., 0.]
        x = np.arange(2)
        y = np.zeros(2)
        self.assertEqual(cost_logistic(p, x, y),
                         cost_abs(p, logistic_p, x, y))
        self.assertEqual(cost2_logistic(p, x, y),
                         cost_square(p, logistic_p, x, y))

    def test_arrhenius(self):
        import numpy as np
        from pyjams.functions import arrhenius, arrhenius_p, cost_arrhenius
        from pyjams.functions import cost2_arrhenius

        T25 = 298.15
        R   = 8.3144621
        p = [1.]
        x = np.arange(2) + 25.
        y = np.zeros(2)
        out = np.exp(np.arange(2) / (T25 * R * (np.arange(2)+T25)))
        self.assertEqual(list(arrhenius(x, *p)), list(out))
        self.assertEqual(list(arrhenius_p(x, p)), list(out))
        self.assertEqual(cost_arrhenius(p, x, y), np.sum(out))
        self.assertEqual(cost2_arrhenius(p, x, y), np.sum(out**2))

    def test_f1x(self):
        import numpy as np
        from pyjams.functions import f1x, f1x_p, cost_f1x, cost2_f1x

        p = [0., 2.]
        x = np.arange(2) + 1.
        y = np.zeros(2)
        out = 2. / x
        self.assertEqual(list(f1x(x, *p)), list(out))
        self.assertEqual(list(f1x_p(x, p)), list(out))
        self.assertEqual(cost_f1x(p, x, y), np.sum(out))
        self.assertEqual(cost2_f1x(p, x, y), np.sum(out**2))

    def test_fexp(self):
        import numpy as np
        from pyjams.functions import fexp, fexp_p, cost_fexp, cost2_fexp

        p = [0., 1., 1.]
        x = np.arange(2)
        y = np.zeros(2)
        out = np.exp(x)
        self.assertEqual(list(fexp(x, *p)), list(out))
        self.assertEqual(list(fexp_p(x, p)), list(out))
        self.assertEqual(cost_fexp(p, x, y), np.sum(out))
        self.assertEqual(cost2_fexp(p, x, y), np.sum(out**2))

    def test_fgauss(self):
        import numpy as np
        from pyjams.functions import gauss, gauss_p, cost_gauss, cost2_gauss

        p = [1., 1.]
        x = np.arange(2)
        y = np.zeros(2)
        out = 1. / np.sqrt(2.*np.pi) * np.exp(-np.abs(x-1.)/2.)
        self.assertEqual(list(gauss(x, *p)), list(out))
        self.assertEqual(list(gauss_p(x, p)), list(out))
        self.assertEqual(cost_gauss(p, x, y), np.sum(out))
        self.assertEqual(cost2_gauss(p, x, y), np.sum(out**2))

    def test_lasslop(self):
        import numpy as np
        from pyjams.functions import lasslop, lasslop_p, cost_lasslop
        from pyjams.functions import cost2_lasslop

        p = [1., 1., 1., 1.]
        Rg = np.arange(2) * 1000.
        et = np.arange(2)
        VPD = np.arange(2) * 10. + 1000.
        y = np.zeros(2)
        beta = np.array([1., np.exp(-10.)])
        out = et - beta*Rg/(Rg+beta)
        self.assertEqual(list(lasslop(Rg, et, VPD, *p)), list(out))
        self.assertEqual(list(lasslop_p(Rg, et, VPD, p)), list(out))
        self.assertEqual(cost_lasslop(p, Rg, et, VPD, y), np.sum(out))
        self.assertEqual(cost2_lasslop(p, Rg, et, VPD, y), np.sum(out**2))

    def test_line(self):
        import numpy as np
        from pyjams.functions import line, line_p, cost_line, cost2_line

        p = [0., 1.]
        x = np.arange(2)
        y = np.zeros(2)
        out = np.arange(2)
        self.assertEqual(list(line(x, *p)), list(out))
        self.assertEqual(list(line_p(x, p)), list(out))
        self.assertEqual(cost_line(p, x, y), np.sum(out))
        self.assertEqual(cost2_line(p, x, y), np.sum(out**2))

    def test_line0(self):
        import numpy as np
        from pyjams.functions import line0, line0_p, cost_line0, cost2_line0

        p = [1.]
        x = np.arange(2)
        y = np.zeros(2)
        out = np.arange(2)
        self.assertEqual(list(line0(x, *p)), list(out))
        self.assertEqual(list(line0_p(x, p)), list(out))
        self.assertEqual(cost_line0(p, x, y), np.sum(out))
        self.assertEqual(cost2_line0(p, x, y), np.sum(out**2))

    def test_lloyd_fix(self):
        import numpy as np
        from pyjams.functions import lloyd_fix, lloyd_fix_p, cost_lloyd_fix
        from pyjams.functions import cost2_lloyd_fix

        p = [1., 1.]
        x = np.arange(2) + 273.15
        y = np.zeros(2)
        out = np.exp(1./56.02 - 1./(x-227.13))
        self.assertEqual(list(lloyd_fix(x, *p)), list(out))
        self.assertEqual(list(lloyd_fix_p(x, p)), list(out))
        self.assertEqual(cost_lloyd_fix(p, x, y), np.sum(out))
        self.assertEqual(cost2_lloyd_fix(p, x, y), np.sum(out**2))

    def test_lloyd_only_rref(self):
        import numpy as np
        from pyjams.functions import lloyd_only_rref, lloyd_only_rref_p
        from pyjams.functions import cost_lloyd_only_rref
        from pyjams.functions import cost2_lloyd_only_rref

        p = [2.]
        x = np.arange(2)
        y = np.zeros(2)
        out = 2. * np.arange(2)
        self.assertEqual(list(lloyd_only_rref(x, *p)), list(out))
        self.assertEqual(list(lloyd_only_rref_p(x, p)), list(out))
        self.assertEqual(cost_lloyd_only_rref(p, x, y), np.sum(out))
        self.assertEqual(cost2_lloyd_only_rref(p, x, y), np.sum(out**2))

    def test_logistic(self):
        import numpy as np
        from pyjams.functions import logistic, logistic_p, cost_logistic
        from pyjams.functions import cost2_logistic

        p = [1., 1., 0.]
        x = np.arange(2)
        y = np.zeros(2)
        out = 1. / (1. + np.exp(-x))
        self.assertEqual(list(logistic(x, *p)), list(out))
        self.assertEqual(list(logistic_p(x, p)), list(out))
        self.assertEqual(cost_logistic(p, x, y), np.sum(out))
        self.assertEqual(cost2_logistic(p, x, y), np.sum(out**2))

    def test_logistic_offset(self):
        import numpy as np
        from pyjams.functions import logistic_offset, logistic_offset_p
        from pyjams.functions import cost_logistic_offset
        from pyjams.functions import cost2_logistic_offset

        p = [1., 1., 0., 1.]
        x = np.arange(2)
        y = np.zeros(2)
        out = 1. / (1. + np.exp(-x)) + 1.
        self.assertEqual(list(logistic_offset(x, *p)), list(out))
        self.assertEqual(list(logistic_offset_p(x, p)), list(out))
        self.assertEqual(cost_logistic_offset(p, x, y), np.sum(out))
        self.assertEqual(cost2_logistic_offset(p, x, y), np.sum(out**2))

    def test_logistic2_offset(self):
        import numpy as np
        from pyjams.functions import logistic2_offset, logistic2_offset_p
        from pyjams.functions import cost_logistic2_offset
        from pyjams.functions import cost2_logistic2_offset

        p = [1., 1., 0., 1., 2., 0., 1.]
        x = np.arange(2)
        y = np.zeros(2)
        out = 1. / (1. + np.exp(-x)) - 1. / (1. + np.exp(-2.*x)) + 1.
        self.assertEqual(list(logistic2_offset(x, *p)), list(out))
        self.assertEqual(list(logistic2_offset_p(x, p)), list(out))
        self.assertEqual(cost_logistic2_offset(p, x, y), np.sum(out))
        self.assertEqual(cost2_logistic2_offset(p, x, y), np.sum(out**2))

    def test_poly(self):
        import numpy as np
        from pyjams.functions import poly, poly_p, cost_poly, cost2_poly

        p = [0., 1., 2., 3]
        x = np.arange(2)
        y = np.zeros(2)
        out = 0. + 1. * x + 2. * x**2 + 3. * x**3
        self.assertEqual(list(poly(x, *p)), list(out))
        self.assertEqual(list(poly_p(x, p)), list(out))
        self.assertEqual(cost_poly(p, x, y), np.sum(out))
        self.assertEqual(cost2_poly(p, x, y), np.sum(out**2))

    def test_sabx(self):
        import numpy as np
        from pyjams.functions import sabx, sabx_p, cost_sabx, cost2_sabx

        p = [0., 2.]
        x = np.arange(2) + 1.
        y = np.zeros(2)
        out = np.sqrt(2. / x)
        self.assertEqual(list(sabx(x, *p)), list(out))
        self.assertEqual(list(sabx_p(x, p)), list(out))
        self.assertEqual(cost_sabx(p, x, y), np.sum(out))
        self.assertEqual(cost2_sabx(p, x, y), np.sum(out**2))

    def test_see(self):
        import numpy as np
        from pyjams.functions import see, see_p, cost_see, cost2_see

        p = [1., 1., 2.]
        x = np.arange(2) + 0.5
        y = np.zeros(2)
        out = np.array([0., 0.25])
        self.assertEqual(list(see(x, *p)), list(out))
        self.assertEqual(list(see_p(x, p)), list(out))
        self.assertEqual(cost_see(p, x, y), np.sum(out))
        self.assertEqual(cost2_see(p, x, y), np.sum(out**2))


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
"""
This is the unittest for ncio module.

python -m unittest -v tests/test_ncio.py
python -m pytest --cov=pyjams --cov-report term-missing -v tests/test_ncio.py

"""
import unittest


def _flatten(itr):
    import numpy as np
    fitr = np.array(itr).flatten()
    if len(fitr) == 0:
        return list(fitr)
    else:
        if isinstance(fitr[0], str):
            return [ i for i in fitr ]
        else:
            return [ i if np.isfinite(i) else np.finfo(float).max
                     for i in fitr ]


class TestNcio(unittest.TestCase):
    """
    Tests for module ncio
    """

    def test_copy_file(self):
        import os
        import numpy as np
        import pyjams.ncio as ncio
        from pyjams import ncinfo, ncread

        ncfile = 'tests/test_readnetcdf.nc'
        ofile = 'test_ncio_copy_file.nc'

        # full copy
        ncio.copy_file(ncfile, ofile)
        fout = ncinfo(ofile, variables=True)
        fsoll = ncinfo(ncfile, variables=True)
        self.assertEqual(fout, fsoll)

        fout = ncinfo(ofile, long_names=True)
        fsoll = ncinfo(ncfile, long_names=True)
        self.assertEqual(fout, fsoll)

        fout = ncinfo(ofile, units=True)
        fsoll = ncinfo(ncfile, units=True)
        self.assertEqual(fout, fsoll)

        fout = ncread(ofile, var='is1')
        fsoll = ncread(ncfile, var='is1')
        self.assertEqual(_flatten(fout), _flatten(fsoll))

        # removevar
        ncio.copy_file(ncfile, ofile, removevar=('is1',))
        fout = ncinfo(ofile, variables=True)
        fsoll = ncinfo(ncfile, variables=True)
        fsoll.remove('is1')
        self.assertEqual(fout, fsoll)

        ncio.copy_file(ncfile, ofile, removevar=['is1', 'is2'])
        fout = ncinfo(ofile, variables=True)
        fsoll = ncinfo(ncfile, variables=True)
        fsoll.remove('is1')
        fsoll.remove('is2')
        self.assertEqual(fout, fsoll)

        # renamevar
        ncio.copy_file(ncfile, ofile, renamevar={'is1': 'is3'})
        fout = ncinfo(ofile, variables=True)
        fsoll = ncinfo(ncfile, variables=True)
        fsoll[fsoll.index('is1')] = 'is3'
        self.assertEqual(fout, fsoll)

        # replacevar
        isout = np.full((2, 4), 3.)
        ncio.copy_file(ncfile, ofile,
                       replacevar={'is1': {'is3': isout}})
        fout = ncinfo(ofile, variables=True)
        fsoll = ncinfo(ncfile, variables=True)
        fsoll[fsoll.index('is1')] = 'is3'
        self.assertEqual(fout, fsoll)

        fout = ncread(ofile, var='is3')
        fsoll = isout
        self.assertEqual(_flatten(fout), _flatten(fsoll))

        # replacevar, replaceatt
        isout = np.full((2, 4), 3.)
        ncio.copy_file(ncfile, ofile,
                       replacevar={'is1': {'is3': isout}},
                       replaceatt={'is2': {'long_name': 'twos'},
                                   'is3': {'long_name': 'threes',
                                           'units': 'arbitrary'}})
        fout = ncinfo(ofile, variables=True)
        fsoll = ncinfo(ncfile, variables=True)
        fsoll[fsoll.index('is1')] = 'is3'
        self.assertEqual(fout, fsoll)

        fout = ncread(ofile, var='is3')
        fsoll = isout
        self.assertEqual(fout.shape, fsoll.shape)
        self.assertEqual(_flatten(fout), _flatten(fsoll))

        att = ncinfo(ofile, var='is2', attributes=True)
        fout = att['long_name']
        fsoll = 'twos'
        self.assertEqual(fout, fsoll)

        att = ncinfo(ofile, var='is3', attributes=True)
        fout = att['long_name']
        fsoll = 'threes'
        self.assertEqual(fout, fsoll)
        fout = att['units']
        fsoll = 'arbitrary'
        self.assertEqual(fout, fsoll)

        # renamevar, noclose
        isout = np.full((2, 4), 3.)
        fo = ncio.copy_file(ncfile, ofile, renamevar={'is1': 'is3'},
                            noclose=True)
        fo.close()
        fout = ncread(ofile, var='is3')
        fsoll = ncread(ncfile, var='is1')
        self.assertEqual(_flatten(fout), _flatten(fsoll))

        fo = ncio.copy_file(ncfile, ofile, renamevar={'is1': 'is3'},
                            noclose=True)
        ovar = fo.variables['is3']
        ovar[:] = isout
        fo.close()
        fout = ncread(ofile, var='is3')
        fsoll = isout
        self.assertEqual(fout.shape, fsoll.shape)
        self.assertEqual(_flatten(fout), _flatten(fsoll))

        # clean up
        if os.path.exists(ofile):
            os.remove(ofile)

    def test_get_fill_value_for_dtype(self):
        import numpy as np
        import netCDF4 as nc
        import pyjams.ncio as ncio
        import warnings

        itypes = [np.int8, np.int16, np.int32, np.int64,
                  np.uint8, np.uint16, np.uint32, np.uint64,
                  np.float32, np.float64]
        otypes = ['i1', 'i2', 'i4', 'i8',
                  'u1', 'u2', 'u4', 'u8',
                  'f4', 'f8']

        for ii, nn in enumerate(itypes):
            out = np.array([1, 2, 3], dtype=nn)
            fout = ncio.get_fill_value_for_dtype(out.dtype)
            fsoll = nc.default_fillvals[otypes[ii]]
            assert fout == fsoll

        # warning unknown type
        with warnings.catch_warnings(record=True) as w:
            nn = np.float128
            out = np.array([1, 2, 3], dtype=nn)
            fout = ncio.get_fill_value_for_dtype(out.dtype)
            assert len(w) > 0

    def test_set_output_filename(self):
        import pyjams.ncio as ncio

        # full copy
        fout = ncio.set_output_filename('in.nc', '-no_patch')
        assert fout == 'in-no_patch.nc'

        fout = ncio.set_output_filename('in.nc', '.nop')
        assert fout == 'in.nop.nc'

        fout = ncio.set_output_filename('in.nc', '')
        assert fout == 'in.nc'

    def test_dimensions(self):
        import os
        import numpy as np
        import netCDF4 as nc
        import pyjams.ncio as ncio
        from pyjams import ncinfo, ncread

        ifile = 'tests/test_ncio.nc'
        ofile = 'test_ncio_dimensions.nc'

        # remove and rename dimension, land is the last dimension
        vtime = 'time'
        vland = 'land'
        vpatch = 'patch'
        opatch = 'opatch'
        fi = nc.Dataset(ifile, 'r')
        if 'file_format' in dir(fi):
            fo = nc.Dataset(ofile, 'w', format=fi.file_format)
        else:
            fo = nc.Dataset(ofile, 'w', format='NETCDF4')
        ntime = fi.dimensions[vtime].size
        ncio.copy_global_attributes(fi, fo,
                                    add={'history': 'removedim/renamedim'},
                                    remove=['CABLE_input_file'])
        ncio.copy_dimensions(fi, fo, removedim=[vland],
                             renamedim={vpatch: opatch})
        ncio.create_variables(fi, fo, time=False, timedim=vtime, fill=True,
                              removedim=vland, renamedim={vpatch: opatch})
        ncio.create_variables(fi, fo, time=True, timedim=vtime, fill=True,
                              removedim=[vland], renamedim={vpatch: opatch})
        for ivar in fi.variables.values():
            if vtime not in ivar.dimensions:
                ovar = fo.variables[ivar.name]
                if vland in ivar.dimensions:
                    ovar[:] = ivar[:].mean(axis=-1)
                else:
                    ovar[:] = ivar[:]
        for tt in range(ntime):
            for ivar in fi.variables.values():
                if vtime in ivar.dimensions:
                    ovar = fo.variables[ivar.name]
                    if vland in ivar.dimensions:
                        ovar[tt, ...] = ivar[tt, ...].mean(axis=-1)
                    else:
                        if ivar.ndim == 1:
                            ovar[tt] = ivar[tt]
                        else:
                            ovar[tt, ...] = ivar[tt, ...]
        fi.close()
        fo.close()

        fout = ncinfo(ofile, variables=True)
        fsoll = ncinfo(ifile, variables=True)
        self.assertEqual(fout, fsoll)

        for vv in ['patchfrac', 'albsoil', 'ratecp', 'SoilMoist']:
            fout = list(ncinfo(ofile, vv, dims=True))
            fsoll = list(ncinfo(ifile, vv, dims=True))
            fsoll = fsoll[:-1]
            if vpatch in fsoll:
                fsoll[fsoll.index(vpatch)] = opatch
            self.assertEqual(fout, fsoll)
            assert opatch in fout
            assert vpatch not in fout

        fout = ncread(ofile, var='SoilTemp')
        fsoll = ncread(ifile, var='SoilTemp')
        self.assertEqual(_flatten(fout), _flatten(fsoll.mean(axis=-1)))

        # changedim, land is the last dimension
        vtime = 'time'
        vland = 'land'
        fi = nc.Dataset(ifile, 'r')
        if 'file_format' in dir(fi):
            fo = nc.Dataset(ofile, 'w', format=fi.file_format)
        else:
            fo = nc.Dataset(ofile, 'w', format='NETCDF4')
        ntime = fi.dimensions[vtime].size
        nland = fi.dimensions[vland].size
        ncio.copy_global_attributes(fi, fo,
                                    add={'history': 'changedim'})
        ncio.copy_dimensions(fi, fo, removedim=None, changedim={vland: 2*nland})
        ncio.create_variables(fi, fo, time=False, timedim=vtime, fill=None,
                              chunksizes=False)
        ncio.create_variables(fi, fo, time=True, timedim=vtime, fill=None,
                              chunksizes=False)
        for ivar in fi.variables.values():
            if vtime not in ivar.dimensions:
                ovar = fo.variables[ivar.name]
                if vland in ivar.dimensions:
                    ovar[:] = np.append(ivar[:], ivar[:], axis=-1)
                else:
                    ovar[:] = ivar[:]
        for tt in range(ntime):
            for ivar in fi.variables.values():
                if vtime in ivar.dimensions:
                    ovar = fo.variables[ivar.name]
                    if vland in ivar.dimensions:
                        ovar[tt, ...] = np.append(ivar[tt, ...], ivar[tt, ...],
                                                  axis=-1)
                    else:
                        if ivar.ndim == 1:
                            ovar[tt] = ivar[tt]
                        else:
                            ovar[tt, ...] = ivar[tt, ...]
        fi.close()
        fo.close()

        fout = ncinfo(ofile, variables=True)
        fsoll = ncinfo(ifile, variables=True)
        self.assertEqual(fout, fsoll)

        for vv in ['patchfrac', 'albsoil', 'ratecp', 'SoilMoist']:
            fout = ncinfo(ofile, vv, shape=True)
            fsoll = ncinfo(ifile, vv, shape=True)
            self.assertEqual(fout[:-1], fsoll[:-1])
            assert fout[-1] == 2*fsoll[-1]

        fout = ncread(ofile, var='SoilTemp')
        fsoll = ncread(ifile, var='SoilTemp')
        fsoll  = np.append(fsoll, fsoll, axis=-1)
        self.assertEqual(_flatten(fout), _flatten(fsoll))

        # adddim, land to fake x and y
        vtime = 'time'
        vland = 'land'
        fi = nc.Dataset(ifile, 'r')
        if 'file_format' in dir(fi):
            fo = nc.Dataset(ofile, 'w', format=fi.file_format)
        else:
            fo = nc.Dataset(ofile, 'w', format='NETCDF4')
        ntime = fi.dimensions[vtime].size
        nland = fi.dimensions[vland].size
        ncio.copy_global_attributes(fi, fo,
                                    add={'history': 'adddim'})
        ncio.copy_dimensions(fi, fo, removedim=['land'],
                             adddim={'xland': nland, 'yland': nland})
        ncio.create_variables(fi, fo, time=False, timedim=vtime, fill=True,
                              izip=False, chunksizes=False,
                              replacedim={vland: ('yland', 'xland')})
        ncio.create_variables(fi, fo, time=True, timedim=vtime, fill=True,
                              izip=True, chunksizes=False,
                              replacedim={vland: ('yland', 'xland')})
        for ivar in fi.variables.values():
            if vtime not in ivar.dimensions:
                ovar = fo.variables[ivar.name]
                if vland in ivar.dimensions:
                    if ivar.ndim == 1:
                        ovar[:] = np.tile(ivar[:], (nland, 1))
                    else:
                        np.tile(ivar[:], nland)
                else:
                    ovar[:] = ivar[:]
        for tt in range(ntime):
            for ivar in fi.variables.values():
                if vtime in ivar.dimensions:
                    ovar = fo.variables[ivar.name]
                    if vland in ivar.dimensions:
                        ovar[tt, ...] = np.tile(ivar[tt, ...], nland)
                    else:
                        if ivar.ndim == 1:
                            ovar[tt] = ivar[tt]
                        else:
                            ovar[tt, ...] = ivar[tt, ...]
        fi.close()
        fo.close()

        fout = ncinfo(ofile, variables=True)
        fsoll = ncinfo(ifile, variables=True)
        self.assertEqual(fout, fsoll)

        for vv in ['patchfrac', 'albsoil', 'ratecp', 'SoilMoist']:
            fout = list(ncinfo(ofile, vv, shape=True))
            fsoll = list(ncinfo(ifile, vv, shape=True))
            fsoll.append(fsoll[-1])
            self.assertEqual(fout, fsoll)

        fout = ncread(ofile, var='SoilTemp')
        fsoll = ncread(ifile, var='SoilTemp')
        fsoll  = np.tile(fsoll, fsoll.shape[-1])
        self.assertEqual(_flatten(fout), _flatten(fsoll))

        # clean up
        if os.path.exists(ofile):
            os.remove(ofile)

    def test_variables(self):
        import os
        import numpy as np
        import netCDF4 as nc
        import pyjams.ncio as ncio
        from pyjams import ncinfo, ncread

        ifile = 'tests/test_ncio.nc'
        ofile = 'test_ncio_variables.nc'

        # remove and rename variables
        vtime = 'time'
        rmvar = ['Cplant']
        mvvar = {'Csoil': 'CarbonSoil'}
        fi = nc.Dataset(ifile, 'r')
        if 'file_format' in dir(fi):
            fo = nc.Dataset(ofile, 'w', format=fi.file_format)
        else:
            fo = nc.Dataset(ofile, 'w', format='NETCDF4')
        ntime = fi.dimensions[vtime].size
        ncio.copy_global_attributes(fi, fo,
                                    add={'history': 'removevar/renamevar',
                                         'extra': 'Add what you want'},
                                    remove=['CABLE_input_file'])
        ncio.copy_dimensions(fi, fo)
        ncio.create_variables(fi, fo, fill=-9999.,
                              removevar=rmvar, renamevar=mvvar)
        ncio.copy_variables(fi, fo, time=False, timedim=vtime,
                            removevar=rmvar, renamevar=mvvar)
        ncio.copy_variables(fi, fo, time=True, timedim=vtime,
                            removevar=rmvar, renamevar=mvvar)
        fi.close()
        fo.close()

        fout = list(ncinfo(ofile, variables=True))
        fsoll = list(ncinfo(ifile, variables=True))
        fsoll.remove(rmvar[0])
        for rr in mvvar:
            fsoll[fsoll.index(rr)] = mvvar[rr]
        self.assertEqual(fout, fsoll)

        # create new variable
        vtime = 'time'
        fi = nc.Dataset(ifile, 'r')
        if 'file_format' in dir(fi):
            fo = nc.Dataset(ofile, 'w', format=fi.file_format)
        else:
            fo = nc.Dataset(ofile, 'w', format='NETCDF4')
        ntime = fi.dimensions[vtime].size
        ncio.copy_global_attributes(fi, fo,
                                    add={'history': 'Create new var'})
        ncio.copy_dimensions(fi, fo)
        ncio.create_variables(fi, fo)
        dtype = fi.variables['SWdown'].dtype
        dimensions = fi.variables['SWdown'].dimensions
        shape = fi.variables['SWdown'].shape
        chunk = list(shape)
        chunk[0] = 1
        odict = {'name': 'test1',
                 'dtype': dtype,
                 'standard_name': 'test1',
                 'long_name': '1st test variable',
                 'dimensions': dimensions}
        test1 = ncio.create_new_variable(
            odict, fo, izip=True, fill=True)
        odict = {'name': 'test2',
                 'dtype': dtype,
                 'standard_name': 'test2',
                 'long_name': '2nd test variable',
                 'dimensions': dimensions}
        test2 = ncio.create_new_variable(
            odict, fo, izip=False, fill=True)
        odict = {'name': 'test3',
                 'dtype': dtype,
                 'standard_name': 'test3',
                 'long_name': '3rd test variable',
                 'dimensions': dimensions}
        test3 = ncio.create_new_variable(
            odict, fo, izip=True, fill=False)
        odict = {'name': 'test4',
                 'dtype': dtype,
                 'standard_name': 'test4',
                 'long_name': '4th test variable',
                 'dimensions': dimensions,
                 'chunksizes': chunk}
        test4 = ncio.create_new_variable(
            odict, fo, izip=True, fill=None)
        odict = {'name': 'test5',
                 'dtype': dtype,
                 'standard_name': 'test5',
                 'long_name': '5th test variable',
                 'dimensions': dimensions}
        test5 = ncio.create_new_variable(
            odict, fo, izip=True, fill=-9999.)
        odict = {'name': 'test6',
                 'dtype': dtype,
                 'standard_name': 'test6',
                 'long_name': '6th test variable',
                 'dimensions': dimensions,
                 'chunksizes': chunk}
        test6 = ncio.create_new_variable(
            odict, fo, izip=True, fill=None, chunksizes=False)
        ncio.copy_variables(fi, fo, time=False, timedim=vtime)
        ncio.copy_variables(fi, fo, time=True, timedim=vtime)
        test1[:] = 1.
        test2[:] = 2.
        test3[:] = 3.
        test4[:] = 4.
        test5[:] = 5.
        test5[::2] = -9999.
        test6[:] = 6.
        fi.close()
        fo.close()

        fout = list(ncinfo(ofile, variables=True))
        assert 'test1' in fout
        assert 'test2' in fout
        assert 'test3' in fout
        assert 'test4' in fout
        assert 'test5' in fout
        assert 'test6' in fout
        fout = [ vv for vv in fout if not vv.startswith('test') ]
        fsoll = list(ncinfo(ifile, variables=True))
        self.assertEqual(fout, fsoll)

        for iv in range(1, 7):
            fout = ncinfo(ofile, f'test{iv}', shape=True)
            fsoll = shape
            self.assertEqual(fout, fsoll)
            fout = ncinfo(ofile, f'test{iv}', dims=True)
            fsoll = dimensions
            self.assertEqual(fout, fsoll)
            fout = ncread(ofile, var=f'test{iv}')
            assert np.all(fout == float(iv))

        # clean up
        if os.path.exists(ofile):
            os.remove(ofile)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
"""
This is the unittest for mad module.

python -m unittest -v tests/test_mad.py
python -m pytest --cov=pyjams --cov-report term-missing -v tests/test_mad.py

"""
import unittest
from numpy.ma import masked


def _same(ist, soll):
    # xor
    thesame = all([ ist[i] != (not soll[i]) for i in range(len(soll)) ])
    return thesame


def _same_masked(ist, soll):
    istlist = list(ist)
    bout = True
    for i in range(len(soll)):
        if (istlist[i] is masked) or (soll[i] is masked):
            bout &= (istlist[i] is masked) and (soll[i] is masked)
        else:
            bout &= istlist[i] == soll[i]
    return bout


class TestMad(unittest.TestCase):
    """
    Tests for mad.py
    """

    def test_mad(self):
        import numpy as np
        from pyjams import mad

        y = np.array([-0.25, 0.68, 0.94, 1.15, 2.26, 2.35, 2.37, 2.40, 2.47,
                      2.54, 2.62, 2.64, 2.90, 2.92, 2.92, 2.93, 3.21, 3.26,
                      3.30, 3.59, 3.68, 4.30, 4.64, 5.34, 5.42, 8.01])

        # raw data
        ist = mad(y)
        soll = [False, False, False, False, False, False, False, False, False,
                False, False, False, False, False, False, False, False, False,
                False, False, False, False, False, False, False, False]
        assert _same(ist, soll)

        ist = mad(y, z=4)
        soll = [False, False, False, False, False, False, False, False, False,
                False, False, False, False, False, False, False, False, False,
                False, False, False, False, False, False, False, True]
        assert _same(ist, soll)

        ist = mad(y, z=3)
        soll = [True, False, False, False, False, False, False, False, False,
                False, False, False, False, False, False, False, False, False,
                False, False, False, False, False, False, True, True]
        assert _same(ist, soll)

        # 1st derivatives
        ist = mad(y, z=4, deriv=1)
        soll = [True, False, False, True, False, False, False, False, False,
                False, False, False, False, False, False, False, False, False,
                False, False, True, False, True, False, True]
        assert _same(ist, soll)

        # 2nd derivatives
        ist = mad(y, z=4, deriv=2)
        soll = [False, False, False, False, False, False, False, False, False,
                False, False, False, False, False, False, False, False, False,
                False, False, False, False, False, True]
        assert _same(ist, soll)

        # direct use
        ist = np.ma.array(y, mask=mad(y, z=4))
        soll = [-0.25, 0.68, 0.94, 1.15, 2.26, 2.35, 2.37, 2.4, 2.47, 2.54,
                2.62, 2.64, 2.9, 2.92, 2.92, 2.93, 3.21, 3.26, 3.3, 3.59, 3.68,
                4.3, 4.64, 5.34, 5.42, masked]
        self.assertEqual(list(np.ma.around(ist, 2)), soll)

        # several dimensions
        yy = np.transpose(np.array([y, y]))
        ist = np.transpose(mad(yy, z=4))
        soll = [[False, False, False, False, False, False, False, False, False,
                 False, False, False, False, False, False, False, False, False,
                 False, False, False, False, False, False, False, True],
                [False, False, False, False, False, False, False, False, False,
                 False, False, False, False, False, False, False, False, False,
                 False, False, False, False, False, False, False, True]]
        ist = list(np.array(ist).flatten())
        soll = list(np.array(soll).flatten())
        assert _same(ist, soll)

        yyy = np.transpose(np.array([y, y, y]))
        ist = np.transpose(mad(yyy, z=3))
        soll = [[True, False, False, False, False, False, False, False, False,
                 False, False, False, False, False, False, False, False,
                 False, False, False, False, False, False, False, True, True],
                [True, False, False, False, False, False, False, False, False,
                 False, False, False, False, False, False, False, False,
                 False, False, False, False, False, False, False, True, True],
                [True, False, False, False, False, False, False, False, False,
                 False, False, False, False, False, False, False, False,
                 False, False, False, False, False, False, False, True, True]]
        ist = list(np.array(ist).flatten())
        soll = list(np.array(soll).flatten())
        assert _same(ist, soll)

        # masked arrays
        my = np.ma.array(y, mask=np.zeros(y.shape))
        my.mask[-1] = True
        ist = mad(my, z=4)
        soll = [True, False, False, False, False, False, False, False, False,
                False, False, False, False, False, False, False, False, False,
                False, False, False, False, False, False, False, masked]
        assert _same_masked(ist, soll)

        ist = mad(my, z=3)
        soll = [True, False, False, False, False, False, False, False, False,
                False, False, False, False, False, False, False, False, False,
                False, False, False, False, False, True, True, masked]
        assert _same_masked(ist, soll)

        my = np.ma.array(y, mask=np.ones(y.shape))
        ist = mad(my, z=4)
        soll = [True, True, True, True, True, True, True, True, True,
                True, True, True, True, True, True, True, True, True,
                True, True, True, True, True, True, True, True]
        assert _same(ist, soll)

        # arrays with NaNs
        ny = y.copy()
        ny[-1] = np.nan
        ist = mad(ny, z=4)
        soll = [True, False, False, False, False, False, False, False, False,
                False, False, False, False, False, False, False, False, False,
                False, False, False, False, False, False, False, False]
        assert _same(ist, soll)

        ist = mad(ny, z=3)
        soll = [True, False, False, False, False, False, False, False, False,
                False, False, False, False, False, False, False, False, False,
                False, False, False, False, False, True, True, False]
        assert _same(ist, soll)

        # all Nans
        ny = y.copy()
        ny[:] = np.nan
        ist = mad(ny, z=4)
        soll = [False, False, False, False, False, False, False, False, False,
                False, False, False, False, False, False, False, False, False,
                False, False, False, False, False, False, False, False]
        assert _same(ist, soll)

        # exclude zeros
        zy = y.copy()
        zy[1] = 0.
        ist = mad(zy, z=3)
        soll = [True, True, False, False, False, False, False, False, False,
                False, False, False, False, False, False, False, False, False,
                False, False, False, False, False, False, True, True]
        assert _same(ist, soll)

        ist = mad(zy, z=3, nozero=True)
        soll = [True, False, False, False, False, False, False, False, False,
                False, False, False, False, False, False, False, False, False,
                False, False, False, False, False, False, True, True]
        assert _same(ist, soll)

        # errors
        # deriv > 2
        self.assertRaises(ValueError, mad, y, deriv=3)
        # ndim > 2
        y3 = np.transpose(np.array([y, y]).reshape(len(y)//2, 2, 2))
        self.assertRaises(ValueError, mad, y3)


if __name__ == "__main__":
    unittest.main()
